import unittest
import numpy as np
import pymoto as pym
import numpy.testing as npt
from scipy.sparse import csc_matrix


class DynamicMatrix(pym.Module):
    alpha = 0.5
    beta = 0.5

    def _response(self, K, M, omega):
        return K + 1j * omega * (self.alpha * M + self.beta * K) - omega ** 2 * M

    def _sensitivity(self, dZ):
        K, M, omega = [s.state for s in self.sig_in]
        dK = np.real(dZ) - (omega * self.beta) * np.imag(dZ)
        dM = (-omega ** 2) * np.real(dZ) - (omega * self.alpha) * np.imag(dZ)
        dZrM = np.real(dZ).contract(M)
        dZiK = np.imag(dZ).contract(K)
        dZiM = np.imag(dZ).contract(M)
        domega = -self.beta * dZiK - self.alpha * dZiM - 2 * omega * dZrM
        return dK, dM, domega


class TestLinSolveModuleSparse(unittest.TestCase):
    # # ------------- Symmetric -------------
    def test_symmetric_real_compliance2d(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 10  # Number of elements
        dom = pym.DomainDefinition(N, N)
        sx = pym.Signal('x', np.random.rand(dom.nel))
        fixed_nodes = dom.get_nodenumber(0, np.arange(0, N+1))
        bc = np.concatenate((fixed_nodes*2, fixed_nodes*2+1))
        # Setup different rhs types
        iforce_x = dom.get_nodenumber(N, np.arange(0, N + 1)) * 2  # Force in x-direction
        iforce_y = dom.get_nodenumber(N, np.arange(0, N + 1)) * 2 + 1  # Force in y-direction

        force_vecs = dict()

        # Single force
        f = np.zeros(dom.nnodes*2)
        f[iforce_x] = 1.0
        force_vecs['single_real'] = f

        # Multiple rhs
        f = np.zeros((dom.nnodes * 2, 2))
        f[iforce_x, 0] = 1.0
        f[iforce_y, 1] = 1.0
        force_vecs['multiple_real'] = f

        for k, f in force_vecs.items():
            with self.subTest(f"RHS-{k}"):
                sf = pym.Signal('f', f)

                fn = pym.Network()
                sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), dom, bc=bc))
                su = fn.append(pym.LinSolve([sK, sf], pym.Signal('u')))

                fn.response()

                self.assertTrue(np.allclose(sK.state@su.state, sf.state))  # Check residual
                # Check finite difference
                # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
                def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
                pym.finite_difference(fn, [sx, sf], su, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_symmetric_real_compliance3d(self):
        """ Test symmetric real sparse matrix (compliance in 3D)"""
        N = 3  # Number of elements
        dom = pym.DomainDefinition(N, N, N)
        sx = pym.Signal('x', np.random.rand(dom.nel))
        jfix, kfix = np.meshgrid(np.arange(0, N+1), np.arange(0, N+1), indexing='ij')
        fixed_nodes = dom.get_nodenumber(0, jfix, kfix).flatten()
        bc = np.concatenate((fixed_nodes*3, fixed_nodes*3+1, fixed_nodes*3+2))
        iforce = dom.get_nodenumber(N, np.arange(0, N+1), np.arange(0, N+1))*3 + 1
        sf = pym.Signal('f', np.zeros(dom.nnodes*3))
        sf.state[iforce] = 1.0

        fn = pym.Network()
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), dom, bc=bc))
        su = fn.append(pym.LinSolve([sK, sf], pym.Signal('u')))

        fn.response()

        self.assertTrue(np.allclose(sK.state@su.state, sf.state))  # Check residual
        # Check finite difference
        # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=2e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sf], su, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_symmetric_complex_dyncompliance2d(self):
        """ Test symmetric complex sparse matrix (dynamic compliance in 2D)"""
        N = 5  # Number of elements
        dom = pym.DomainDefinition(N, N)
        sx = pym.Signal('x', np.random.rand(dom.nel))
        fixed_nodes = dom.get_nodenumber(0, np.arange(0, N+1))
        bc = np.concatenate((fixed_nodes*2, fixed_nodes*2+1))
        iforce = dom.get_nodenumber(N, np.arange(0, N+1))*2 + 1
        sf = pym.Signal('f', np.zeros(dom.nnodes*2))
        sf.state[iforce] = 1.0

        sOmega = pym.Signal('omega', 0.1)
        fn = pym.Network()
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), dom, bc=bc))
        sM = fn.append(pym.AssembleMass(sx, pym.Signal('M'), dom, bc=bc))
        sZ = fn.append(DynamicMatrix([sK, sM, sOmega], pym.Signal('Z')))

        su = fn.append(pym.LinSolve([sZ, sf], pym.Signal('u')))

        fn.response()

        # spspla.eigsh(sK.state, M=sM.state, k=6, sigma=0.0)

        self.assertTrue(np.allclose(sZ.state@su.state, sf.state))  # Check residual
        # Check finite difference
        # def tfn(x0, dx, df_an, df_fd): np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5)
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sf, sOmega], su, test_fn=tfn, dx=1e-7, tol=1e-4, verbose=False)


class TestAssemblyAddValues(unittest.TestCase):
    def test_finite_difference(self):
        N = 2
        # Set up the domain
        domain = pym.DomainDefinition(N, N)
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_x = dofs_left[0::2]
        dofs_left_y = dofs_left[1::2]
        dof_input = dofs_left_y[0]  # Input dofs for mechanism
        dof_output = dofs_left_y[-1]  # Output dofs for mechanism

        prescribed_dofs = np.union1d(dofs_left_x, dofs_right)

        # Setup rhs for two loadcases
        f = np.zeros(domain.nnodes * 2, dtype=float)
        f[dof_input] = 1.0

        # Initial design
        sx = pym.Signal('x', np.random.rand(domain.nel))
        signal_force = pym.Signal('f', state=f)
        # Setup optimization problem
        network = pym.Network()

        # Assembly
        istiff = np.array([dof_input, dof_output])
        sstiff = np.array([10.0, 10.0])

        K_const = csc_matrix((sstiff, (istiff, istiff)), shape=(domain.nnodes * 2, domain.nnodes * 2))
        signal_stiffness = network.append(pym.AssembleStiffness(sx, domain=domain, bc=prescribed_dofs, add_constant=K_const))
        su = network.append(pym.LinSolve([signal_stiffness, signal_force], pym.Signal('u')))
        sc = network.append(pym.EinSum(su, expression='i->'))
        network.response()

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))

        pym.finite_difference(network, [sx], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_added_stiffness_on_ground(self):
        N = 2
        # Set up the domain
        domain = pym.DomainDefinition(N, N)
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_x = dofs_left[0::2]
        dofs_left_y = dofs_left[1::2]
        dof_input = dofs_left_y[0]  # Input dofs for mechanism
        dof_output = dofs_left_y[-1]  # Output dofs for mechanism

        prescribed_dofs = np.union1d(dofs_left_x, dofs_right)

        # Setup rhs for two loadcases
        f = np.zeros(domain.nnodes * 2, dtype=float)
        f[dof_input] = 1.0

        # Initial design
        sx = pym.Signal('x', np.random.rand(domain.nel))
        signal_force = pym.Signal('f', state=f)
        # Setup optimization problem
        network = pym.Network()

        # Assembly
        istiff = np.array([dof_input, dof_output])
        sstiff = np.array([10.0, 10.0])

        K_const = csc_matrix((sstiff, (istiff, istiff)), shape=(domain.nnodes * 2, domain.nnodes * 2))
        signal_stiffness = network.append(
            pym.AssembleStiffness(sx, domain=domain, bc=prescribed_dofs, add_constant=K_const))
        su = network.append(pym.LinSolve([signal_stiffness, signal_force], pym.Signal('u')))
        sc = network.append(pym.EinSum(su, expression='i->'))

        network.response()

        network2 = pym.Network()
        # Assembly
        istiff = np.array([dof_input, dof_output, 4, 5])
        sstiff = np.array([10.0, 10.0, 100.0, 100.0])

        K_const1 = csc_matrix((sstiff, (istiff, istiff)), shape=(domain.nnodes * 2, domain.nnodes * 2))
        signal_stiffness = network2.append(
            pym.AssembleStiffness(sx, domain=domain, bc=prescribed_dofs, add_constant=K_const1))
        su = network2.append(pym.LinSolve([signal_stiffness, signal_force], pym.Signal('u')))
        sc2 = network2.append(pym.EinSum(su, expression='i->'))

        network2.response()

        npt.assert_allclose(sc.state, sc2.state)


class TestSystemOfEquations(unittest.TestCase):
    def test_sparse_symmetric_real_compliance2d_single_load(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N=10
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_horizontal = dofs_left[0::2]
        dofs_left_vertical = dofs_left[1::2]

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = np.unique(np.hstack([dofs_left_horizontal, dofs_right, dofs_left_vertical]))
        free_dofs = np.setdiff1d(all_dofs, prescribed_dofs)

        # Setup solution vectors and rhs
        ff = np.zeros_like(free_dofs, dtype=float)
        ff[:] = np.random.rand(len(free_dofs))
        u = np.zeros_like(all_dofs, dtype=float)

        u[dofs_left_vertical] = np.random.rand(len(dofs_left_vertical))
        up = u[prescribed_dofs]

        sff = pym.Signal('ff', ff)
        sup = pym.Signal('up', up)

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.SystemOfEquations([sK, sff, sup], free=free_dofs, prescribed=prescribed_dofs))
        sc = fn.append(pym.EinSum([su[0], su[1]], expression='i,i->'))
        fn.response()
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sff, sup], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_sparse_symmetric_real_compliance2d_single_load_u(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 10
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_horizontal = dofs_left[0::2]
        dofs_left_vertical = dofs_left[1::2]

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = np.unique(np.hstack([dofs_left_horizontal, dofs_right, dofs_left_vertical]))
        free_dofs = np.setdiff1d(all_dofs, prescribed_dofs)

        # Setup solution vectors and rhs
        ff = np.zeros_like(free_dofs, dtype=float)
        ff[:] = np.random.rand(len(free_dofs))
        u = np.zeros_like(all_dofs, dtype=float)

        u[dofs_left_vertical] = np.random.rand(len(dofs_left_vertical))
        up = u[prescribed_dofs]

        sff = pym.Signal('ff', ff)
        sup = pym.Signal('up', up)

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.SystemOfEquations([sK, sff, sup], free=free_dofs, prescribed=prescribed_dofs))
        sc = fn.append(pym.EinSum([su[0]], expression='i->'))
        fn.response()

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))

        pym.finite_difference(fn, [sx, sff, sup], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_sparse_symmetric_real_compliance2d_single_load_f(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 10
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_horizontal = dofs_left[0::2]
        dofs_left_vertical = dofs_left[1::2]

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = np.unique(np.hstack([dofs_left_horizontal, dofs_right, dofs_left_vertical]))
        free_dofs = np.setdiff1d(all_dofs, prescribed_dofs)

        # Setup solution vectors and rhs
        ff = np.zeros_like(free_dofs, dtype=float)
        ff[:] = np.random.rand(len(free_dofs))
        u = np.zeros_like(all_dofs, dtype=float)

        u[dofs_left_vertical] = np.random.rand(len(dofs_left_vertical))
        up = u[prescribed_dofs]

        sff = pym.Signal('ff', ff)
        sup = pym.Signal('up', up)

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.SystemOfEquations([sK, sff, sup], free=free_dofs, prescribed=prescribed_dofs))
        sc = fn.append(pym.EinSum([su[1]], expression='i->'))
        fn.response()

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))

        pym.finite_difference(fn, [sx, sff, sup], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_sparse_symmetric_real_compliance2d_single_multi_load(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N=10
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_horizontal = dofs_left[0::2]
        dofs_left_vertical = dofs_left[1::2]

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = np.unique(np.hstack([dofs_left_horizontal, dofs_right, dofs_left_vertical]))
        free_dofs = np.setdiff1d(all_dofs, prescribed_dofs)

        # Setup solution vectors and rhs
        ff = np.zeros((len(free_dofs), 2), dtype=float)
        ff[:, :] = np.random.rand(np.shape(ff)[0], np.shape(ff)[1])
        u = np.zeros((len(all_dofs), 2), dtype=float)

        u[dofs_left_vertical, 0] = np.random.rand(len(dofs_left_vertical))
        u[dofs_left_horizontal, 1] = np.random.rand(len(dofs_left_vertical))
        up = u[prescribed_dofs, :]

        sff = pym.Signal('ff', ff)
        sup = pym.Signal('up', up)

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.SystemOfEquations([sK, sff, sup], free=free_dofs, prescribed=prescribed_dofs))
        sc1 = fn.append(pym.EinSum([su[0][:, 0], su[1][:, 0]], expression='i,i->'))
        sc2 = fn.append(pym.EinSum([su[0][:, 1], su[1][:, 1]], expression='i,i->'))
        sc = fn.append(pym.MathGeneral([sc1, sc2], expression='inp0 + inp1'))
        fn.response()
        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))
        pym.finite_difference(fn, [sx, sff, sup], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_sparse_symmetric_real_compliance2d_single_multi_load_u(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 10
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_horizontal = dofs_left[0::2]
        dofs_left_vertical = dofs_left[1::2]

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = np.unique(np.hstack([dofs_left_horizontal, dofs_right, dofs_left_vertical]))
        free_dofs = np.setdiff1d(all_dofs, prescribed_dofs)

        # Setup solution vectors and rhs
        ff = np.zeros((len(free_dofs), 2), dtype=float)
        ff[:, :] = np.random.rand(np.shape(ff)[0], np.shape(ff)[1])
        u = np.zeros((len(all_dofs), 2), dtype=float)

        u[dofs_left_vertical, 0] = np.random.rand(len(dofs_left_vertical))
        u[dofs_left_horizontal, 1] = np.random.rand(len(dofs_left_vertical))
        up = u[prescribed_dofs, :]

        sff = pym.Signal('ff', ff)
        sup = pym.Signal('up', up)

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.SystemOfEquations([sK, sff, sup], free=free_dofs, prescribed=prescribed_dofs))
        sc = fn.append(pym.EinSum(su[0], expression='ij->'))
        fn.response()

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))

        pym.finite_difference(fn, [sx, sff, sup], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)

    def test_sparse_symmetric_real_compliance2d_single_multi_load_f(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 10
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_left_horizontal = dofs_left[0::2]
        dofs_left_vertical = dofs_left[1::2]

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = np.unique(np.hstack([dofs_left_horizontal, dofs_right, dofs_left_vertical]))
        free_dofs = np.setdiff1d(all_dofs, prescribed_dofs)

        # Setup solution vectors and rhs
        ff = np.zeros((len(free_dofs), 2), dtype=float)
        ff[:, :] = np.random.rand(np.shape(ff)[0], np.shape(ff)[1])
        u = np.zeros((len(all_dofs), 2), dtype=float)

        u[dofs_left_vertical, 0] = np.random.rand(len(dofs_left_vertical))
        u[dofs_left_horizontal, 1] = np.random.rand(len(dofs_left_vertical))
        up = u[prescribed_dofs, :]

        sff = pym.Signal('ff', ff)
        sup = pym.Signal('up', up)

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.SystemOfEquations([sK, sff, sup], free=free_dofs, prescribed=prescribed_dofs))
        sc = fn.append(pym.EinSum(su[1], expression='ij->'))
        fn.response()

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))

        pym.finite_difference(fn, [sx, sff, sup], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)


if __name__ == '__main__':
    unittest.main()



import numpy as np
from pymoto.utils import _parse_to_list, _concatenate_to_array


def residual(x, y, z, lam, xsi, eta, mu, zet, s, upp, low, P0, P1, Q0, Q1, epsi, a0, a, b, c, d, alfa, beta):
    # upcoming lines determine the left hand sides, i.e. the resiudals of all constraints
    ux1 = upp - x
    xl1 = x - low

    plam = P0 + np.dot(lam, P1)
    qlam = Q0 + np.dot(lam, Q1)
    gvec = np.dot(P1, 1/ux1) + np.dot(Q1, 1/xl1)

    # gradient of approximation function wrt x
    dpsidx = plam / (ux1**2) - qlam / (xl1**2)

    # put all residuals in one line
    return np.concatenate([
        dpsidx - xsi + eta,  # rex [n]
        c + d * y - mu - lam,  # rey [m]
        np.array([a0 - zet - np.dot(a, lam)]),  # rez [1]
        gvec - a * z - y + s - b,  # relam [m]
        xsi * (x - alfa) - epsi,  # rexsi [n]
        eta * (beta - x) - epsi,  # reeta [n]
        mu * y - epsi,  # remu [m]
        np.array([zet * z - epsi]),  # rezet [1]
        lam * s - epsi,  # res [m]
    ])


def subsolv(epsimin, low, upp, alfa, beta, P, Q, a0, a, b, c, d):
    """ This function subsolv solves the MMA subproblem
    minimize   SUM[ p0j/(uppj-xj) + q0j/(xj-lowj) ] + a0*z +
             + SUM[ ci*yi + 0.5*di*(yi)^2 ],
    subject to SUM[ pij/(uppj-xj) + qij/(xj-lowj) ] - ai*z - yi <= bi,
               alfaj <=  xj <=  betaj,  yi >= 0,  z >= 0.
    Input:  m, n, low, upp, alfa, beta, p0, q0, P, Q, a0, a, b, c, d.
    Output: xmma,ymma,zmma, slack variables and Lagrange multiplers.
    """

    n, m = len(alfa), len(a)
    epsi = 1.0
    maxittt = 400
    x = 0.5 * (alfa + beta)
    y = np.ones(m)
    z = 1.0
    lam = np.ones(m)
    GG = np.empty((m, n))
    xsi = np.maximum((1.0 / (x - alfa)), 1)
    eta = np.maximum((1.0 / (beta - x)), 1)
    mu = np.maximum(1, 0.5 * c)
    zet = 1.0
    s = np.ones(m)
    bb = np.empty(m+1)
    AA = np.empty((m+1, m+1))

    P0 = np.ascontiguousarray(P[0, :])
    Q0 = np.ascontiguousarray(Q[0, :])
    P1 = np.ascontiguousarray(P[1:, :])
    Q1 = np.ascontiguousarray(Q[1:, :])

    itera = 0
    while epsi > epsimin:
        # main loop + 1
        itera = itera + 1

        # upcoming lines determine the left hand sides, i.e. the resiudals of all constraints
        residu = residual(x, y, z, lam, xsi, eta, mu, zet, s, upp, low, P0, P1, Q0, Q1, epsi, a0, a, b, c, d, alfa, beta)
        residunorm = np.linalg.norm(residu)
        residumax = np.max(np.abs(residu))

        ittt = 0
        # the algorithm is terminated when the maximum residual has become smaller than 0.9*epsilon
        # and epsilon has become sufficiently small (and not too many iterations are used)
        while residumax > 0.9 * epsi and ittt < maxittt:
            ittt = ittt + 1

            # Newton's method: first create the variable steps

            # precalculations for PSIjj (or diagx)
            ux1 = upp - x
            xl1 = x - low
            ux2 = ux1 ** 2
            xl2 = xl1 ** 2
            ux3 = ux1 * ux2
            xl3 = xl1 * xl2

            uxinv1 = 1.0 / ux1
            xlinv1 = 1.0 / xl1
            uxinv2 = 1.0 / ux2
            xlinv2 = 1.0 / xl2

            plam = P0 + np.dot(lam, P1)
            qlam = Q0 + np.dot(lam, Q1)
            gvec = np.dot(P1, uxinv1) + np.dot(Q1, xlinv1)

            # CG is an m x n matrix with values equal to partial derivative of constraints wrt variables
            GG[:, :] = P1 * uxinv2 - Q1 * xlinv2

            # derivative of PSI wrt x
            dpsidx = plam / ux2 - qlam / xl2

            # calculation of right hand sides dx, dy, dz, dlam
            delx = dpsidx - epsi / (x - alfa) + epsi / (beta - x)
            dely = c + d * y - lam - epsi / y
            delz = a0 - np.dot(a, lam) - epsi / z
            dellam = gvec - a * z - y - b + epsi / lam

            # calculation of diagonal matrices Dx Dy Dlam
            diagx = 2 * (plam / ux3 + qlam / xl3) + xsi / (x - alfa) + eta / (beta - x)
            diagy = d + mu / y

            diaglam = s / lam
            diaglamyi = diaglam + 1.0 / diagy

            # different options depending on the number of constraints
            # considering the fact I will probably not use local constraints I removed the option

            # normally here is a statement if m < n
            bb[:-1] = dellam + dely / diagy - np.dot(GG, (delx / diagx))
            bb[-1] = delz

            AA[:-1, :-1] = np.diag(diaglamyi) + np.dot((GG / diagx), GG.T)
            AA[-1, :-1] = a
            AA[:-1, -1] = a
            AA[-1, -1] = -zet/z
            # solve system for delta lambda and delta z
            solut = np.linalg.solve(AA, bb)

            # solution of delta vars
            dlam = solut[0:m]
            dz = solut[m]
            dx = -delx / diagx - np.dot(dlam, GG) / diagx
            dy = -dely / diagy + dlam / diagy
            dxsi = -xsi + epsi / (x - alfa) - (xsi * dx) / (x - alfa)
            deta = -eta + epsi / (beta - x) + (eta * dx) / (beta - x)
            dmu = -mu + epsi / y - (mu * dy) / y
            dzet = -zet + epsi / z - zet * dz / z
            ds = -s + epsi / lam - (s * dlam) / lam

            # calculate the step size
            stmy = -1.01*np.min(dy/y)
            stmz = -1.01 * dz / z
            stmlam = -1.01*np.min(dlam / lam)
            stmxsi = -1.01*np.min(dxsi / xsi)
            stmeta = -1.01 * np.min(deta / eta)
            stmmu = -1.01 * np.min(dmu / mu)
            stmzet = -1.01 * dzet / zet
            stms = -1.01 * np.min(ds / s)
            stmxx = max(stmy, stmz, stmlam, stmxsi, stmeta, stmmu, stmzet, stms)

            # put variables and accompanying changes in alist
            stmalfa = -1.01 * np.min(dx / (x - alfa))
            stmbeta = 1.01 * np.max(dx / (beta - x))

            # Initial step size
            steg = 1.0 / max(stmalfa, stmbeta, stmxx, 1.0)

            # set old variables
            xold = x.copy()
            yold = y.copy()
            zold = z
            lamold = lam.copy()
            xsiold = xsi.copy()
            etaold = eta.copy()
            muold = mu.copy()
            zetold = zet
            sold = s.copy()

            # Do linesearch
            itto = 0
            while itto < maxittt:
                # Find new set of variables with stepsize
                x[:] = xold + steg * dx
                y[:] = yold + steg * dy
                z = zold + steg * dz
                lam[:] = lamold + steg * dlam
                xsi[:] = xsiold + steg * dxsi
                eta[:] = etaold + steg * deta
                mu[:] = muold + steg * dmu
                zet = zetold + steg * dzet
                s[:] = sold + steg * ds

                residu = residual(x, y, z, lam, xsi, eta, mu, zet, s, upp, low, P0, P1, Q0, Q1, epsi, a0, a, b, c, d, alfa, beta)
                if np.linalg.norm(residu) < residunorm:
                    break
                itto += 1
                steg /= 2  # Reduce stepsize

            residunorm = np.linalg.norm(residu)
            residumax = np.max(np.abs(residu))

        if ittt > maxittt - 2:
            print(f"MMA Subsolver: itt = {ittt}, at epsi = {epsi}")
        # decrease epsilon with factor 10
        epsi /= 10

    # ## END OF SUBSOLVE
    return x, y, z, lam, xsi, eta, mu, zet, s


class MMA:
    """
    Block for the MMA algorithm
    The design variables are set by keyword <variables> accepting a list of variables.
    The responses are set by keyword <responses> accepting a list of signals.
    If none are given, the internal sig_in and sig_out are used.

    Args:
        function: The Network defining the optimization problem
        variables: The Signals defining the design variables
        responses: A list of Signals, where the first is to be minimized and the others are constraints.

    Keyword Args:
        tolx: Stopping criterium for relative design change
        tolf: Stopping criterium for relative objective change
        maxit: Maximum number of iteration
        move: Move limit on relative variable change per iteration
        xmin: Minimum design variable (can be a vector)
        xmax: Maximum design variable (can be a vector)
        fn_callback: A function that is called just before calling the response() in each iteration
        verbosity: 0 - No prints, 1 - Only convergence message, 2 - Convergence and iteration info, 3 - Extended info

    """

    def __init__(self, function, variables, responses, tolx=1e-4, tolf=0.0, move=0.1, maxit=100, xmin=0.0, xmax=1.0, fn_callback=None, verbosity=0, **kwargs):
        self.funbl = function
        self.verbosity = verbosity

        self.variables = _parse_to_list(variables)
        self.responses = _parse_to_list(responses)

        self.iter = 0

        # Convergence options
        self.tolX = tolx
        self.tolf = tolf
        self.maxIt = maxit

        # Operational options
        self.xmax = xmax
        self.xmin = xmin
        self.move = move

        self.pijconst = kwargs.get("pijconst", 1e-3)

        self.a0 = kwargs.get("a0", 1.0)

        self.epsimin = kwargs.get("epsimin", 1e-7)  # Or 1e-7 ?? witout sqrt(m+n) or 1e-9
        self.raa0 = kwargs.get("raa0", 1e-5)

        self.cCoef = kwargs.get("cCoef", 1e3)  # Svanberg uses 1e3 in example? Old code had 1e7

        # Not used
        self.dxmin = kwargs.get("dxmin", 1e-5)

        self.albefa = kwargs.get("albefa", 0.1)
        self.asyinit = kwargs.get("asyinit", 0.5)
        self.asyincr = kwargs.get("asyincr", 1.2)
        self.asydecr = kwargs.get("asydecr", 0.7)
        self.asybound = kwargs.get("asybound", 10.0)

        self.ittomax = kwargs.get("ittomax", 400)

        self.iterinitial = kwargs.get("iterinitial", 2.5)

        self.fn_callback = fn_callback

        # Numbers
        self.n = None  # len(x0)
        self.dx = None
        self.xold1 = None
        self.xold2 = None
        self.low = None
        self.upp = None
        self.offset = None

        # Setting up for constriants
        self.m = len(self.responses) - 1
        self.a = np.zeros(self.m)
        self.c = self.cCoef * np.ones(self.m)
        self.d = np.ones(self.m)
        self.gold1 = np.zeros(self.m + 1)
        self.gold2 = self.gold1.copy()
        self.rho = self.raa0 * np.ones(self.m + 1)

    def response(self):
        change = 1

        # Save initial state
        xval, self.cumlens = _concatenate_to_array([s.state for s in self.variables])
        self.n = len(xval)

        # Set outer bounds
        if not hasattr(self.xmin, '__len__'):
            self.xmin = self.xmin * np.ones_like(xval)
        elif len(self.xmin) == len(self.variables):
            xminvals = self.xmin
            self.xmin = np.zeros_like(xval)
            for i in range(len(xminvals)):
                self.xmin[self.cumlens[i]:self.cumlens[i+1]] = xminvals[i]

        if len(self.xmin) != self.n:
            raise RuntimeError(
                "Length of the xmin vector not correct ({} != {})".format(len(self.xmin), self.n))

        if not hasattr(self.xmax, '__len__'):
            self.xmax = self.xmax * np.ones_like(xval)
        elif len(self.xmax) == len(self.variables):
            xmaxvals = self.xmax
            self.xmax = np.zeros_like(xval)
            for i in range(len(xmaxvals)):
                self.xmax[self.cumlens[i]:self.cumlens[i + 1]] = xmaxvals[i]

        if len(self.xmax) != self.n:
            raise RuntimeError("Length of the xmax vector not correct ({} != {})".format(len(self.xmax), self.n))

        # Set movelimit in case of multiple
        if hasattr(self.move, '__len__'):
            if len(self.move) == len(self.variables):
                movevals = self.move
                self.move = np.zeros_like(xval)
                for i in range(len(movevals)):
                    self.move[self.cumlens[i]:self.cumlens[i + 1]] = movevals[i]
            elif len(self.move) != self.n:
                raise RuntimeError("Length of the move vector not correct ({} != {})".format(len(self.move), self.n))

        fcur = 0.0
        while self.iter < self.maxIt:
            # Reset all signals in function block
            self.funbl.reset()

            # Set the new states
            for i, s in enumerate(self.variables):
                if self.cumlens[i+1]-self.cumlens[i] == 1:
                    try:
                        s.state[:] = xval[self.cumlens[i]]
                    except TypeError:
                        s.state = xval[self.cumlens[i]]
                else:
                    s.state[:] = xval[self.cumlens[i]:self.cumlens[i+1]]

            if self.fn_callback is not None:
                self.fn_callback()

            # Calculate response
            self.funbl.response()

            # Update the states
            for i, s in enumerate(self.variables):
                if self.cumlens[i+1]-self.cumlens[i] == 1:
                    try:
                        xval[self.cumlens[i]] = s.state[:]
                    except (TypeError, IndexError):
                        xval[self.cumlens[i]] = s.state
                else:
                    xval[self.cumlens[i]:self.cumlens[i+1]] = s.state[:]

            # Save response
            f = ()
            for s in self.responses:
                f += (s.state, )

            # Check function change convergence criterion
            fprev, fcur = fcur, self.responses[0].state
            rel_fchange = abs(fcur-fprev)/abs(fcur)
            if rel_fchange < self.tolf:
                if self.verbosity >= 1:
                    print(f"MMA converged: Relative function change |Δf|/|f| ({rel_fchange}) below tolerance ({self.tolf})")
                break

            # Calculate and save sensitivities
            df = ()
            for i, s_out in enumerate(self.responses):
                for s in self.responses:
                    s.reset()

                s_out.sensitivity = s_out.state*0 + 1.0

                self.funbl.sensitivity()

                sens_list = []
                for v in self.variables:
                    sens_list.append(v.sensitivity if v.sensitivity is not None else 0*v.state)
                dff, _ = _concatenate_to_array(sens_list)
                df += (dff, )

                # Reset sensitivities for the next response
                self.funbl.reset()

            # Display info on variables
            if self.verbosity >= 3:
                for i, s in enumerate(self.variables):
                    isscal = self.cumlens[i + 1] - self.cumlens[i] == 1
                    msg = "{0:>10s} = ".format(s.tag)
                    if isscal:
                        try:
                            msg += "         {0: .3e}         ".format(s.state)
                        except TypeError:
                            msg += "         {0: .3e}         ".format(s.state[0])
                    else:
                        msg += "[{0: .3e} ... {1: .3e}] ".format(min(s.state), max(s.state))
                    for j, s_out in enumerate(self.responses):
                        msg += "| {0:>10s}/{1:10s} = ".format("d"+s_out.tag, "d"+s.tag)
                        if isscal:
                            msg += "         {0: .3e}         ".format(df[j][self.cumlens[i]])
                        else:
                            msg += "[{0: .3e} ... {1: .3e}] ".format(min(df[j][self.cumlens[i]:self.cumlens[i+1]]), max(df[j][self.cumlens[i]:self.cumlens[i+1]]))
                    print(msg)

            self.iter += 1
            xnew, change = self.mmasub(xval.copy(), np.hstack(f), np.vstack(df))

            # Stopping criteria on step size
            rel_stepsize = np.linalg.norm((xval - xnew)/self.dx) / np.linalg.norm(xval/self.dx)
            if rel_stepsize < self.tolX:
                if self.verbosity >= 1:
                    print(f"MMA converged: Relative stepsize |Δx|/|x| ({rel_stepsize}) below tolerance ({self.tolX})")
                break

            xval = xnew

    def mmasub(self, xval, g, dg):
        if self.dx is None:
            self.dx = self.xmax - self.xmin
        if self.offset is None:
            self.offset = self.asyinit * np.ones(self.n)

        #      Minimize  f_0(x) + a_0*z + sum( c_i*y_i + 0.5*d_i*(y_i)^2 )
        #    subject to  f_i(x) - a_i*z - y_i <= 0,  i = 1,...,m
        #                xmin_j <= x_j <= xmax_j,    j = 1,...,n
        #                z >= 0,   y_i >= 0,         i = 1,...,m
        # *** INPUT:
        #
        #   m    = The number of general constraints.
        #   n    = The number of variables x_j.
        #  iter  = Current iteration number ( =1 the first time mmasub is called).
        #  xval  = Column vector with the current values of the variables x_j.
        #  xmin  = Column vector with the lower bounds for the variables x_j.
        #  xmax  = Column vector with the upper bounds for the variables x_j.
        #  xold1 = xval, one iteration ago (provided that iter>1).
        #  xold2 = xval, two iterations ago (provided that iter>2).
        #  f0val = The value of the objective function f_0 at xval.
        #  df0dx = Column vector with the derivatives of the objective function
        #          f_0 with respect to the variables x_j, calculated at xval.
        #  fval  = Column vector with the values of the constraint functions f_i,
        #          calculated at xval.
        #  dfdx  = (m x n)-matrix with the derivatives of the constraint functions
        #          f_i with respect to the variables x_j, calculated at xval.
        #          dfdx(i,j) = the derivative of f_i with respect to x_j.
        #  low   = Column vector with the lower asymptotes from the previous
        #          iteration (provided that iter>1).
        #  upp   = Column vector with the upper asymptotes from the previous
        #          iteration (provided that iter>1).
        #  a0    = The constants a_0 in the term a_0*z.
        #  a     = Column vector with the constants a_i in the terms a_i*z.
        #  c     = Column vector with the constants c_i in the terms c_i*y_i.
        #  d     = Column vector with the constants d_i in the terms 0.5*d_i*(y_i)^2.
        #

        # *** OUTPUT:
        #
        #  xmma  = Column vector with the optimal values of the variables x_j
        #          in the current MMA subproblem.
        #  ymma  = Column vector with the optimal values of the variables y_i
        #          in the current MMA subproblem.
        #  zmma  = Scalar with the optimal value of the variable z
        #          in the current MMA subproblem.
        #  lam   = Lagrange multipliers for the m general MMA constraints.
        #  xsi   = Lagrange multipliers for the n constraints alfa_j - x_j <= 0.
        #  eta   = Lagrange multipliers for the n constraints x_j - beta_j <= 0.
        #   mu   = Lagrange multipliers for the m constraints -y_i <= 0.
        #  zet   = Lagrange multiplier for the single constraint -z <= 0.
        #   s    = Slack variables for the m general MMA constraints.
        #  low   = Column vector with the lower asymptotes, calculated and used
        #          in the current MMA subproblem.
        #  upp   = Column vector with the upper asymptotes, calculated and used
        #          in the current MMA subproblem.

        # # ASYMPTOTES
        # Calculation of the asymptotes low and upp :
        # For iter = 1,2 the asymptotes are fixed depending on asyinit
        if self.xold1 is not None and self.xold2 is not None:
            # depending on if the signs of xval - xold and xold - xold2 are opposite, indicating an oscillation
            # in the variable xi
            # if the signs are equal the asymptotes are slowing down the convergence and should be relaxed

            # check for oscillations in variables
            # if zzz positive no oscillations, if negative --> oscillations
            zzz = (xval - self.xold1) * (self.xold1 - self.xold2)
            # decrease those variables that are oscillating equal to asydecr
            self.offset[zzz > 0] *= self.asyincr
            self.offset[zzz < 0] *= self.asydecr

            # check with minimum and maximum bounds of asymptotes, as they cannot be to close or far from the variable
            # give boundaries for upper and lower asymptotes
            self.offset = np.clip(self.offset, 1/(self.asybound**2), self.asybound)

        # Update asymptotes
        shift = self.offset * self.dx
        self.low = xval - shift
        self.upp = xval + shift

        # # VARIABLE BOUNDS
        # Calculation of the bounds alfa and beta :
        # use albefa to limit the maximum change of variables wrt the lower and upper asymptotes
        # as it should remain within both asymptotes
        zzl1 = self.low + self.albefa * shift
        # use movelimit to limit the maximum change of variables
        zzl2 = xval - self.move * self.dx
        # minimum variable bounds
        alfa = np.maximum.reduce([zzl1, zzl2, self.xmin])

        zzu1 = self.upp - self.albefa * shift
        zzu2 = xval + self.move * self.dx
        # maximum variable bounds
        beta = np.minimum.reduce([zzu1, zzu2, self.xmax])

        # # APPROXIMATE CONVEX SEPARABLE FUNCTIONS
        # Calculations of p0, q0, P, Q and b.
        # calculate the constant factor in calculations of pij and qij
        dx2 = shift**2
        P = dx2 * np.maximum(+dg, 0)
        Q = dx2 * np.maximum(-dg, 0)

        rhs = np.dot(P, 1 / shift) + np.dot(Q, 1 / shift) - g
        b = rhs[1:]

        # Solving the subproblem by a primal-dual Newton method
        epsimin_scaled = self.epsimin*np.sqrt(self.m + self.n)
        xmma, ymma, zmma, lam, xsi, eta, mu, zet, s = subsolv(epsimin_scaled, self.low, self.upp, alfa, beta, P, Q, self.a0, self.a, b, self.c, self.d)

        self.gold2, self.gold1 = self.gold1, g.copy()
        self.xold2, self.xold1 = self.xold1, xval.copy()
        change = np.average(abs(xval - xmma))

        if self.verbosity >= 2:
            msgs = ["g{0:d}({1:s}): {2:+.4e}".format(i, s.tag, g[i]) for i, s in enumerate(self.responses)]
            print("It. {0: 4d}, {1}".format(self.iter, ", ".join(msgs)))

        if self.verbosity >=3:
            # Print changes
            printstr = "Changes: "
            for i, s in enumerate(self.variables):
                isscal = self.cumlens[i + 1] - self.cumlens[i] == 1
                if isscal:
                    chg = abs(xval[self.cumlens[i]] - xmma[self.cumlens[i]])
                else:
                    chg = np.average(abs(xval[self.cumlens[i]:self.cumlens[i + 1]] - xmma[self.cumlens[i]:self.cumlens[i + 1]]))

                printstr += "{0:s} = {1:.3e}   ".format("Δ_"+s.tag, chg)
            print(printstr)

        return xmma, change


import unittest
import numpy as np
import pymoto as pym


class TestScaling(unittest.TestCase):
    def test_objective(self):
        sx = pym.Signal('x', 1.0)
        m = pym.Scaling(sx, scaling=105.0)
        m.response()
        self.assertEqual(m.sig_out[0].state, 105.0)
        sx.state = 2.0
        m.response()
        self.assertEqual(m.sig_out[0].state, 210.0)

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-7, atol=1e-5))
        pym.finite_difference(m, test_fn=tfn)

    def test_lower_constraint(self):
        sx = pym.Signal('x', 1.0)
        m = pym.Scaling(sx, scaling=15.0, minval=0.5)
        m.response()
        self.assertLess(m.sig_out[0].state, 0.0)
        sx.state = 0.5
        m.response()
        self.assertEqual(m.sig_out[0].state, 0.0)
        sx.state = 0.4
        m.response()
        self.assertGreater(m.sig_out[0].state, 0.0)
        sx.state = 0.0
        m.response()
        self.assertEqual(m.sig_out[0].state, 15.0)

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-7, atol=1e-5))
        pym.finite_difference(m, test_fn=tfn)

    def test_upper_constraint(self):
        sx = pym.Signal('x', 1.0)
        m = pym.Scaling(sx, scaling=15.0, maxval=0.5)
        m.response()
        self.assertGreater(m.sig_out[0].state, 0.0)
        sx.state = 0.5
        m.response()
        self.assertEqual(m.sig_out[0].state, 0.0)
        sx.state = 0.4
        m.response()
        self.assertLess(m.sig_out[0].state, 0.0)
        sx.state = 1.0
        m.response()
        self.assertEqual(m.sig_out[0].state, 15.0)

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-7, atol=1e-5))
        pym.finite_difference(m, test_fn=tfn)


import unittest
import numpy as np
import pymoto as pym
import numpy.testing as npt


class TestAssembleStiffness(unittest.TestCase):
    def test_FEA_pure_tensile_2d_one_element(self):
        Lx, Ly, Lz = 0.1, 0.2, 0.3
        domain = pym.DomainDefinition(1, 1, unitx=Lx, unity=Ly, unitz=Lz)
        nodidx_left = domain.get_nodenumber(0, np.arange(domain.nely + 1))
        # Fixed at bottom, roller at the top in y-direction
        nodidx_right = domain.get_nodenumber(domain.nelx, np.arange(domain.nely + 1))
        dofidx_left = np.concatenate([nodidx_left*2, np.array([nodidx_left[0]*2 + 1])])

        E, nu = 210e+9, 0.3

        s_x = pym.Signal('x', state=np.ones(domain.nel))

        # Assemble stiffness matrix
        m_K = pym.AssembleStiffness(s_x, domain=domain, bc=dofidx_left, e_modulus=E, poisson_ratio=nu, plane='stress')
        s_K = m_K.sig_out[0]

        m_K.response()
        F = 1.5
        f = np.zeros(domain.nnodes*2)
        f[nodidx_right * 2] = F/nodidx_right.size
        x = np.linalg.solve(s_K.state.toarray(), f)

        # Bottom y displacement should be zero
        npt.assert_allclose(x[nodidx_right[0]*2+1], 0, atol=1e-10)

        # Analytical axial displacement using stiffness k = EA/L
        ux_chk = F * Lx / (E * Ly * Lz)
        npt.assert_allclose(x[nodidx_right*2], ux_chk, rtol=1e-10)

        # Transverse displacement using Poisson's effect
        e_xx = ux_chk / Lx  # Strain in x-direction
        e_yy = - nu * e_xx
        uy_chk = e_yy * Ly
        npt.assert_allclose(x[nodidx_right[1]*2+1], uy_chk, rtol=1e-10)

    def test_FEA_pure_tensile_3d_one_element(self):
        Lx, Ly, Lz = 0.1, 0.2, 0.3
        domain = pym.DomainDefinition(1, 1, 1, unitx=Lx, unity=Ly, unitz=Lz)
        nodidx_left = domain.get_nodenumber(*np.meshgrid(0, range(domain.nely + 1), range(domain.nelz + 1))).flatten()
        # Fixed at (0,0,0), roller in z-direction at (0, 1, 0), roller in y-direction at (0, 0, 1)
        nod_00 = domain.get_nodenumber(0, 0, 0)
        nod_10 = domain.get_nodenumber(0, 1, 0)
        nod_01 = domain.get_nodenumber(0, 0, 1)
        dofidx_left = np.concatenate([nodidx_left * 3, np.array([nod_00, nod_01]) * 3 + 1, np.array([nod_00, nod_10]) * 3 + 2])
        nodidx_right = domain.get_nodenumber(*np.meshgrid(1, range(domain.nely + 1), range(domain.nelz + 1))).flatten()

        E, nu = 210e+9, 0.3

        s_x = pym.Signal('x', state=np.ones(domain.nel))

        # Assemble stiffness matrix
        m_K = pym.AssembleStiffness(s_x, domain=domain, bc=dofidx_left, e_modulus=E, poisson_ratio=nu)
        s_K = m_K.sig_out[0]

        m_K.response()
        F = 1.5e+3
        f = np.zeros(domain.nnodes * 3)
        f[nodidx_right * 3] = F / nodidx_right.size
        x = np.linalg.solve(s_K.state.toarray(), f)

        # y and z displacements at (1, 0, 0) should be zero
        npt.assert_allclose(x[domain.get_nodenumber(1, 0, 0) * 3 + 1], 0, atol=1e-10)
        npt.assert_allclose(x[domain.get_nodenumber(1, 0, 0) * 3 + 2], 0, atol=1e-10)

        # Z displacement at (1, 1, 0) should be zero
        npt.assert_allclose(x[domain.get_nodenumber(1, 1, 0) * 3 + 2], 0, atol=1e-10)

        # Y displacement at (1, 0, 1) should be zero
        npt.assert_allclose(x[domain.get_nodenumber(1, 0, 1) * 3 + 1], 0, atol=1e-10)

        # Analytical axial displacement using stiffness k = EA/L
        ux_chk = F * Lx / (E * Ly * Lz)
        npt.assert_allclose(x[nodidx_right * 3], ux_chk, rtol=1e-10)

        # Transverse displacement using Poisson's effect
        e_xx = ux_chk / Lx  # Strain in x-direction
        e_trans = - nu * e_xx
        uy_chk = e_trans * Ly
        npt.assert_allclose(x[domain.get_nodenumber(1, 1, 0) * 3 + 1], uy_chk, rtol=1e-10)
        npt.assert_allclose(x[domain.get_nodenumber(1, 1, 1) * 3 + 1], uy_chk, rtol=1e-10)

        uz_chk = e_trans * Lz
        npt.assert_allclose(x[domain.get_nodenumber(1, 0, 1) * 3 + 2], uz_chk, rtol=1e-10)
        npt.assert_allclose(x[domain.get_nodenumber(1, 1, 1) * 3 + 2], uz_chk, rtol=1e-10)


import inspect
import pathlib  # For importing files
import sys
import unittest
from math import isclose

import numpy as np
import scipy as sp
import scipy.sparse as spsp
import scipy.sparse.linalg as spspla
from scipy.io import mmread  # For importing files

import pymoto as pym

try:
    import cvxopt
    _has_cvxopt = True
except ImportError:
    sys.stdout.write("Skipping all tests with CVXOPT matrix...\n")
    _has_cvxopt = False

np.random.seed(0)  # Set seed for repeatability


def calculate_eigenvalues(A):
    """ Calculates the eigenvalues of a matrix to check positive-definiteness """
    return sp.linalg.eigvals(A.toarray())


def load_matrix(filen):
    """ Loads a matrix from file """
    curdir = pathlib.Path(__file__).parent.resolve()
    return mmread(pathlib.Path(curdir, "data", filen))


def generate_real_diagonal(N):
    """ Generates a real diagonal matrix of size N """
    return spsp.spdiags(np.random.rand(N), 0, N, N).tocoo()  # All positive


def generate_complex_diagonal(N):
    """ Generates a complex diagonal matrix of size N """
    return spsp.spdiags(np.random.rand(N) + 1j*np.random.rand(N), 0, N, N).tocoo()


def generate_symm_indef():
    """ Generates a real symmetric indefinite matrix """
    A = load_matrix("zenios.mtx.gz")
    sigma = -0.1  # To prevent exact singularity
    As = A - sigma*spsp.eye(A.shape[0])
    # Calculate eigenvalues
    # solver = pym.SolverSparseLU().update(As)
    # opinv = spspla.LinearOperator(A.shape, matvec=solver.solve, rmatvec=solver.adjoint)
    # w, _ = spspla.eigsh(A, sigma=sigma, k=100, OPinv=opinv, ncv=200)]
    return As.tocoo()


def generate_dynamic(complex=False, indef=False):
    """ Generate a dynamic stiffness matrix, either complex or not, and indefinite or not"""
    K = load_matrix("bcsstk26.mtx.gz")
    M = load_matrix("bcsstm26.mtx.gz")
    solver = pym.SolverSparseLU().update(K.tocsc())
    opinv = spspla.LinearOperator(K.shape, matvec=solver.solve, rmatvec=solver.adjoint)
    w, _ = spspla.eigsh(K, M=M, sigma=0, k=100, OPinv=opinv, ncv=200)
    eigfreqs = np.sqrt(w)
    if indef:
        freq = (eigfreqs[0] + eigfreqs[1])/2
    else:
        freq = eigfreqs[0]/2
    if complex:
        Ks = K + freq*1j*(0.01*K + 0.01*M) - freq**2 * M
    else:
        Ks = K - freq**2 * M
    return Ks.tocoo()


def generate_saddlepoint():
    K = load_matrix("bcsstk26.mtx.gz")
    M = load_matrix("bcsstm26.mtx.gz")
    solver = pym.SolverSparseLU().update(K.tocsc())
    opinv = spspla.LinearOperator(K.shape, matvec=solver.solve, rmatvec=solver.adjoint)
    w, V = spspla.eigsh(K, M=M, sigma=0, k=100, OPinv=opinv, ncv=200)
    wi, vi = w[0], V[:, [0]]
    Mv = M@vi

    return spsp.bmat([[K-wi*M, -Mv],
                      [-Mv.T, None]]).tocoo()


""" Perpare the set of matrices we want to test """
# Real-valued matrices
mat_real_spdiag = spsp.spdiags(np.random.rand(100), 0, 100, 100)  # Diagonal (spdiag format)
mat_real_diagonal = generate_real_diagonal(100)  # Diagonal (coo format)
mat_real_symm_pos_def = load_matrix("bcsstk14.mtx.gz")  # Symmetric positive definite
mat_real_symm_pos_def_dynamic = generate_dynamic(complex=False, indef=False)  # Symmetric positive definite
mat_real_symm_indef = generate_symm_indef()  # Symmetric indefinite
mat_real_symm_indef_dynamic = generate_dynamic(complex=False, indef=True)  # Symmetric indefinite
mat_real_symm_saddle = generate_saddlepoint()  # Indefinite saddlepoinnt
mat_real_asymm = load_matrix("impcol_a.mtx.gz")  # Asymmetric

# Complex-valued matrices
mat_complex_spdiag = mat_real_spdiag + 1j*mat_real_spdiag  # Diagonal (spdiag format)
mat_complex_diagonal = generate_complex_diagonal(100)  # Diagonal (coo format)
mat_complex_symm_pos_def_dynamic = generate_dynamic(complex=True, indef=False)  # Symmetric positive definite
mat_complex_symm_indef_dynamic = generate_dynamic(complex=True, indef=True)  # Symmetric indefinite
mat_complex_hermitian_1 = load_matrix("mhd1280b.mtx.gz")  # Hermitian
mat_complex_hermitian_indef = mat_complex_hermitian_1 - spsp.eye(*mat_complex_hermitian_1.shape)
mat_complex_hermitian_pos_def = mat_complex_hermitian_1 + spsp.eye(*mat_complex_hermitian_1.shape)

""" ------------------ TEST UTILITY FUNCTIONS -------------------- """


class TestGenericUtility(unittest.TestCase):
    """ Generic test class for testing functions on matrix properties that return True/False """
    data = []  # The data to test: list of tuples with (matrix, True/False property)
    fn = None  # The test function to use, which return the True/False property

    @classmethod
    def setUpClass(cls) -> None:
        """ Skip test if no function is given """
        if cls.fn is None:
            raise unittest.SkipTest(f"Skipping test {cls}")

    def get_tags(self):
        """ Generate tags for all the matrices to use in messaging"""
        src = inspect.getsource(type(self))
        start = src.find('data')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')

        def parse_str(t: str):
            rmv = ['\n', ' ', '(', ',', 'True', 'False']
            for c in rmv:
                t = t.replace(c, '')
            return t

        return [parse_str(t) for t in src[start:finish].strip(' []').split(')')]

    def test_all_matrices(self):
        """ Test all the matrices """
        tags = self.get_tags()
        for i, dat in enumerate(self.data):
            A = dat[0]
            res = dat[1]
            # Numpy dense matrix
            with self.subTest(msg=f"Dense - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for dense matrix {tags[i]}\n")
                self.assertEqual(self.fn(A.toarray()), res)

            # Scipy COO / spdiag matrix
            with self.subTest(msg=f"{type(A)} - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for {type(A)}-matrix {tags[i]}\n")
                self.assertEqual(self.fn(A), res)

            # Scipy CSR matrix
            with self.subTest(msg=f"CSR - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for CSR-matrix {tags[i]}\n")
                self.assertEqual(self.fn(A.tocsr()), res)

            # Scipy CSC matrix
            with self.subTest(msg=f"CSC - {tags[i]}"):
                # sys.stdout.write(f"Test {self.fn.__name__} for CSC-matrix {tags[i]}\n")
                self.assertEqual(self.fn(A.tocsc()), res)

            # CVXOPT spmatrix
            if _has_cvxopt:
                with self.subTest(msg=f"CVXOPT - {tags[i]}"):
                    # sys.stdout.write(f"Test {self.fn.__name__} for CVXOPT-matrix {tags[i]}\n")
                    if isinstance(A, spsp.coo_matrix):
                        Acvx = cvxopt.spmatrix(A.data, A.row.astype(int), A.col.astype(int))
                    self.assertEqual(self.fn(Acvx), res)


class TestIsComplex(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_complex)
    data = [
        (mat_real_diagonal, False),
        (mat_real_spdiag, False),
        (mat_real_symm_pos_def, False),
        (mat_real_symm_pos_def_dynamic, False),
        (mat_real_symm_indef, False),
        (mat_real_symm_indef_dynamic, False),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_spdiag, True),
        (mat_complex_symm_pos_def_dynamic, True),
        (mat_complex_symm_indef_dynamic, True),
        (mat_complex_hermitian_1, True),
        (mat_complex_hermitian_indef, True),
        (mat_complex_hermitian_pos_def, True),
    ]


class TestIsDiagonal(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_diagonal)
    data = [
        (mat_real_diagonal, True),
        (mat_real_spdiag, True),
        (mat_real_symm_pos_def, False),
        (mat_real_symm_pos_def_dynamic, False),
        (mat_real_symm_indef, False),
        (mat_real_symm_indef_dynamic, False),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_spdiag, True),
        (mat_complex_symm_pos_def_dynamic, False),
        (mat_complex_symm_indef_dynamic, False),
        (mat_complex_hermitian_1, False),
        (mat_complex_hermitian_indef, False),
        (mat_complex_hermitian_pos_def, False),
    ]


class TestIsSymmetric(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_symmetric)
    data = [
        (mat_real_diagonal, True),
        (mat_real_spdiag, True),
        (mat_real_symm_pos_def, True),
        (mat_real_symm_pos_def_dynamic, True),
        (mat_real_symm_indef, True),
        (mat_real_symm_indef_dynamic, True),
        (mat_real_asymm, False),
        (mat_complex_diagonal, True),
        (mat_complex_spdiag, True),
        (mat_complex_symm_pos_def_dynamic, True),
        (mat_complex_symm_indef_dynamic, True),
        (mat_complex_hermitian_1, False),
        (mat_complex_hermitian_indef, False),
        (mat_complex_hermitian_pos_def, False),
    ]


class TestIsHermitian(TestGenericUtility):
    fn = staticmethod(pym.matrix_is_hermitian)
    data = [
        (mat_real_diagonal, True),
        (mat_real_spdiag, True),
        (mat_real_symm_pos_def, True),
        (mat_real_symm_pos_def_dynamic, True),
        (mat_real_symm_indef, True),
        (mat_real_symm_indef_dynamic, True),
        (mat_real_asymm, False),
        (mat_complex_diagonal, False),
        (mat_complex_spdiag, False),
        (mat_complex_symm_pos_def_dynamic, False),
        (mat_complex_symm_indef_dynamic, False),
        (mat_complex_hermitian_1, True),
        (mat_complex_hermitian_indef, True),
        (mat_complex_hermitian_pos_def, True),
    ]


""" ------------------ TEST THE SPARSE SOLVERS -------------------- """
# SOLVERS:
# - scipy/umfpack lu
# - scikit cholmod
# - cvxopt cholmod
# - intel mkl pardiso


class GenericTestSolvers(unittest.TestCase):
    """ Generic test runner for any solver and list of matrices """
    solver = None  # The solver to use
    matrices = []  # The list of matrices the solver should be able to handle

    @classmethod
    def setUpClass(cls) -> None:
        """ Skip test if the solver is not available """
        if cls.solver is None:
            raise unittest.SkipTest(f"No solver given for {cls.__name__}")
        if not cls.solver.defined:
            raise unittest.SkipTest(f"Solver \"{cls.solver.__name__}\" not available, probably lacking third-party libraries")

    def get_tags(self):
        """ Get the matrix tags for messages """
        src = inspect.getsource(type(self))
        start = src.find('matrices')
        start += src[start:].find('=') + 1
        finish = start + src[start:].find(']')
        return [t.strip(' ').replace('\n', '').replace(' ', '') for t in src[start:finish].strip(' []').split(',')]

    def run_solver(self, solver, A, b):
        """ Run the actual test for given solver, matrix and right-hand-side """
        # Reference solution
        x_ref = pym.SolverSparseLU(A).solve(b)
        xadj_ref = pym.SolverSparseLU(A.conj().T).solve(b)

        # Calculate the solution and the adjoint solution
        solver.update(A)
        x = solver.solve(b)
        xadj = solver.adjoint(b)

        # Check residual and solution
        residual = A @ x - b
        r = np.linalg.norm(residual) / np.linalg.norm(b)
        self.assertTrue(r < 1e-4, msg=f"Residual is {r}")
        self.assertTrue(np.allclose(x, x_ref))

        # Check adjoint residual and solution
        residual_adj = (A.conj().T) @ xadj - b
        r_adj = np.linalg.norm(residual_adj) / np.linalg.norm(b)
        self.assertTrue(r_adj < 1e-4, msg=f"Adjoint residual is {r_adj} (normal residual = {r})")
        self.assertTrue(np.allclose(xadj, xadj_ref))

    def run_sparse_tests(self, solver, Acoo, rhs):
        """ Run the sparse tests for different matrix types """
        self.run_solver(solver, Acoo, rhs)
        self.run_solver(solver, Acoo.tocsr(), rhs)
        self.run_solver(solver, Acoo.tocsc(), rhs)

    def test_all_matrices(self):
        """ Run the tests on all given matrices """
        tags = self.get_tags()
        for A, t in zip(self.matrices, tags):
            """ Test for different type of right-hand-sides """
            sys.stdout.write(f"Test \"{self.solver.__name__}\" for matrix \"{t}\"\n")
            N = A.shape[0]
            with self.subTest(msg=f"{t}.real-rhs"):
                b = np.random.rand(N)
                self.run_sparse_tests(self.solver(), A, b)
            with self.subTest(msg=f"{t}.multi-real-rhs"):
                b = np.random.rand(N, 3)
                self.run_sparse_tests(self.solver(), A, b)

            if not pym.matrix_is_complex(A):
                continue

            with self.subTest(msg=f"{t}.complex-rhs"):
                b = np.random.rand(N) + 1j * np.random.rand(N)
                self.run_sparse_tests(self.solver(), A, b)
            with self.subTest(msg=f"{t}.multi-complex-rhs"):
                b = np.random.rand(N, 3) + 1j * np.random.rand(N, 3)
                self.run_sparse_tests(self.solver(), A, b)


class TestSparseDiagonal(GenericTestSolvers):
    solver = pym.SolverDiagonal
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_complex_diagonal,
        mat_complex_spdiag,
    ]


class TestSparseLU(GenericTestSolvers):
    solver = pym.SolverSparseLU
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_real_symm_indef,
        mat_real_symm_indef_dynamic,
        mat_real_asymm,
        mat_real_symm_saddle,
        mat_complex_diagonal,
        mat_complex_spdiag,
        mat_complex_symm_pos_def_dynamic,
        mat_complex_symm_indef_dynamic,
        mat_complex_hermitian_1,
        mat_complex_hermitian_indef,
        mat_complex_hermitian_pos_def,
    ]


class TestCholeskyScikit(GenericTestSolvers):
    # I don't know why, but scikit is able to solve indefinite matrix as well. Maybe they do some LDL inside?
    solver = pym.SolverSparseCholeskyScikit
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_complex_hermitian_indef,
        mat_complex_hermitian_1,
        mat_complex_hermitian_pos_def,
    ]


class TestCholeskyCVXOPT(GenericTestSolvers):
    solver = pym.SolverSparseCholeskyCVXOPT
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_complex_hermitian_1,
        mat_complex_hermitian_pos_def,
    ]


class TestPardiso(GenericTestSolvers):
    solver = pym.SolverSparsePardiso
    matrices = [
        mat_real_diagonal,
        mat_real_spdiag,
        mat_real_symm_pos_def,
        mat_real_symm_pos_def_dynamic,
        mat_real_symm_indef,
        mat_real_symm_indef_dynamic,
        mat_real_symm_saddle,
        mat_real_asymm,
        # mat_complex_diagonal, # PyPardiso does not work for complex matrix yet
        # mat_complex_spdiag,
        # mat_complex_symm_pos_def_dynamic,
        # mat_complex_symm_indef_dynamic,
        # mat_complex_hermitian_1,
    ]


if __name__ == '__main__':
    unittest.main()



from pymoto import Module


class Scaling(Module):
    r""" Scales (scalar) input for different response functions in optimization (objective / constraints).
    This is useful, for instance, for MMA where the objective must be scaled in a certain way for good convergence.

    Objective scaling (`minval` and `maxval` are both undefined):
    :math:`y^{(i)} = s \frac{x^{(i)}}{x^{(0)}}`

    For the constraints, the negative null-form convention is used, which means the constraint is :math:`y(x) \leq 0`.

    Upper limit constraint :math:`x\leq x_\text{max}` (`maxval` is specified):
    :math:`y = s \left( \frac{x}{x_\text{max}} - 1 \right)`

    Lower limit constraint :math:`x\geq x_\text{min}` (`minval` is specified):
    :math:`y = s \left( 1 - \frac{x}{x_\text{min}} \right)`

    Input Signal:
        - ``x``: Unscaled variable :math:`x`

    Output Signal:
        - ``y``: Scaled variable :math:`y`

    Keyword Args:
        scaling: Value :math:`s` to scale with
        minval: Minimum value :math:`x_\text{min}` for negative-null-form constraint
        minval: Maximum value :math:`x_\text{max}` for negative-null-form constraint
    """
    def _prepare(self, scaling: float = 100.0, minval: float = None, maxval: float = None):
        self.minval = minval
        self.maxval = maxval
        self.scaling = scaling
        if self.minval is not None and self.maxval is not None:
            raise RuntimeError("Only one-sided constraints are allowed. Either provide only minval or only maxval.")
        # In case of constraints, initial x-value is not required for scaling-factor
        if self.minval is not None or self.maxval is not None:
            self.sf = self.scaling

    def _response(self, x):
        if not hasattr(self, 'sf'):
            self.sf = self.scaling/x
        if self.minval is not None:
            g = 1 - x/self.minval
        elif self.maxval is not None:
            g = x/self.maxval - 1
        else:
            g = x
        return g * self.sf

    def _sensitivity(self, dy):
        dg = dy * self.sf
        if self.minval is not None:
            return - dg / self.minval
        elif self.maxval is not None:
            return dg / self.maxval
        else:
            return dg




""" Specialized linear algebra modules """
import warnings
from inspect import currentframe, getframeinfo

import numpy as np
import scipy.linalg as spla  # Dense matrix solvers
import scipy.sparse as sps
import scipy.sparse.linalg as spsla

from pymoto import Signal, Module, DyadCarrier, LDAWrapper
from pymoto import SolverDenseLU, SolverDenseLDL, SolverDenseCholesky, SolverDiagonal, SolverDenseQR
from pymoto import SolverSparseLU, SolverSparseCholeskyCVXOPT, SolverSparsePardiso, SolverSparseCholeskyScikit
from pymoto import matrix_is_symmetric, matrix_is_hermitian, matrix_is_diagonal


class StaticCondensation(Module):
    r"""Static condensation of a linear system of equations

    The partitioned system of equations

    :math:`\begin{bmatrix} \mathbf{A}_\text{mm} & \mathbf{A}_\text{ms} \\ \mathbf{A}_\text{sm} & \mathbf{A}_\text{ss}
    \end{bmatrix}
    \begin{bmatrix} \mathbf{x}_\text{m} \\ \mathbf{x}_\text{s} \end{bmatrix} =
    \begin{bmatrix} \mathbf{b}_\text{m} \\ \mathbf{b}_\text{s} \end{bmatrix}
    ,`
    with subscripts ``(m)`` and ``(s)`` referring to the main and secondary dofs, respectively.

    The system is solved in two steps:

    :math:`\begin{aligned}
    \mathbf{A}_\text{ss} \mathbf{x}_\text{sm} &= \mathbf{A}_\text{sm} \\
    \tilde{\mathbf{A}} &= \mathbf{A}_\text{mm} - \mathbf{A}_\text{ms} \mathbf{x}_\text{sm}.
    \end{aligned}`

    Assumptions:
        (i) It is assumed the prescribed DOFs (all dof - main dof - free dof) are prescribed to zero.
        (ii) It is assumed the applied load on the free DOFs is zero; there is no reduced load.

    Implemented by @artofscience (s.koppen@tudelft.nl).

    References:

    Koppen, S., Langelaar, M., & van Keulen, F. (2022).
    Efficient multi-partition topology optimization.
    Computer Methods in Applied Mechanics and Engineering, 393, 114829.
    DOI: https://doi.org/10.1016/j.cma.2022.114829

    Input Signals:
      - ``A`` (`dense or sparse matrix`): The system matrix :math:`\mathbf{A}` of size ``(n, n)``

    Output Signal:
      - ``Ared`` (`dense or sparse matrix`): The reduced system matrix :math:`\tilde{\mathbf{A}}` of size ``(m, m)``

    Args:
        free: The indices corresponding to the free degrees of freedom
        main: The indices corresponding to the main degrees of freedom
        **kwargs: See `pymoto.LinSolve`, as they are directly passed into the `LinSolve` module
    """

    def _prepare(self, main, free, **kwargs):
        self.module_LinSolve = LinSolve([self.sig_in[0], Signal()], **kwargs)
        self.module_LinSolve.use_lda_solver = False
        self.m = main
        self.f = free

    def _response(self, A):
        self.n = np.shape(A)[0]
        self.module_LinSolve.sig_in[0].state = A[self.f, ...][..., self.f]
        self.module_LinSolve.sig_in[1].state = A[self.f, ...][..., self.m].todense()
        self.module_LinSolve.response()
        self.X = self.module_LinSolve.sig_out[0].state
        return A[self.m, ...][..., self.m] - A[self.m, ...][..., self.f] @ self.X

    def _sensitivity(self, dfdB):
        C = np.zeros((self.n, len(self.m)), dtype=float)
        C[self.m, ...] = np.eye(len(self.m))
        C[self.f, ...] = -self.X
        return C @ dfdB @ C.T if isinstance(dfdB, DyadCarrier) else DyadCarrier(list(C.T), list(np.asarray(dfdB @ C.T)))


class SystemOfEquations(Module):
    r""" Solve a partitioned linear system of equations

    The partitioned system of equations

    :math:`\begin{bmatrix} \mathbf{A}_\text{ff} & \mathbf{A}_\text{fp} \\ \mathbf{A}_\text{pf} & \mathbf{A}_\text{pp}
    \end{bmatrix}
    \begin{bmatrix} \mathbf{x}_\text{f} \\ \mathbf{x}_\text{p} \end{bmatrix} =
    \begin{bmatrix} \mathbf{b}_\text{f} \\ \mathbf{b}_\text{p} \end{bmatrix}
    ,`

    which is solved in two steps. First solving for the free unknowns (e.g. displacements or temperatures)
    :math:`\mathbf{x}_f` and then calculating the rhs for the prescribed unknowns (e.g. reaction forces or heat flux):

    :math:`\begin{aligned}
    \mathbf{A}_\text{ff} \mathbf{x}_\text{f} &= \mathbf{b}_\text{f} - \mathbf{A}_\text{fp} \mathbf{x}_\text{p} \\
    \mathbf{b}_\text{p} &= \mathbf{A}_\text{pf}\mathbf{x}_\text{f} + \mathbf{A}_\text{pp} \mathbf{x}_\text{p}.
    \end{aligned}`

    References:
        Koppen, S., Langelaar, M., & van Keulen, F. (2022).
        Efficient multi-partition topology optimization.
        Computer Methods in Applied Mechanics and Engineering, 393, 114829.
        DOI: https://doi.org/10.1016/j.cma.2022.114829

    Input Signals:
      - ``A`` (`dense or sparse matrix`): The system matrix :math:`\mathbf{A}` of size ``(n, n)``
      - ``b_f`` (`vector`): applied load vector of size ``(f)`` or block-vector of size ``(f, Nrhs)``
      - ``x_p`` (`vector`): prescribed state vector of size ``(p)`` or block-vector of size ``(p, Nrhs)``

    Output Signal:
      - ``x`` (`vector`): state vector of size ``(n)`` or block-vector of size ``(n, Nrhs)``
      - ``b`` (`vector`): load vector of size ``(n)`` or block-vector of size ``(n, Nrhs)``

    Keyword Args:
        free: The indices corresponding to the free degrees of freedom at which :math:`f_\text{f}` is given
        prescribed: The indices corresponding to the prescibed degrees of freedom at which :math:`x_\text{p}` is given
        **kwargs: See `pymoto.LinSolve`, as they are directly passed into the `LinSolve` module
    """

    def _prepare(self, free=None, prescribed=None, **kwargs):
        self.module_LinSolve = LinSolve([self.sig_in[0], Signal()], **kwargs)
        self.f = free
        self.p = prescribed
        assert self.p is not None or self.f is not None, "Either prescribed or free indices must be provided"

    def _response(self, A, bf, xp):
        assert bf.shape[0] + xp.shape[0] == A.shape[0], "Dimensions of applied force and displacement must match matrix"
        assert bf.ndim == xp.ndim, "Number of loadcases for applied force and displacement must match"
        self.n = np.shape(A)[0]
        self.dim = xp.ndim

        if self.f is None:
            all_dofs = np.arange(self.n)
            self.f = np.setdiff1d(all_dofs, self.p)
        if self.p is None:
            all_dofs = np.arange(self.n)
            self.p = np.setdiff1d(all_dofs, self.f)
        assert self.f.size + self.p.size == self.n, "Size of free and prescribed indices must match the matrix size"

        # create empty output
        self.x = np.zeros((self.n, *bf.shape[1:]), dtype=complex) if np.iscomplexobj(A) else np.zeros(
            (self.n, *bf.shape[1:]), dtype=float)
        self.x[self.p, ...] = xp

        b = np.zeros_like(self.x)
        b[self.f, ...] = bf

        # partitioning
        Aff = A[self.f, :][:, self.f]
        self.Afp = A[self.f, :][:, self.p]
        self.App = A[self.p, :][:, self.p]

        # solve
        self.module_LinSolve.sig_in[0].state = Aff
        self.module_LinSolve.sig_in[1].state = bf - self.Afp * xp
        self.module_LinSolve.response()
        xf = self.module_LinSolve.sig_out[0].state

        # set output
        self.x[self.f, ...] = xf
        b[self.p, ...] = self.Afp.T * xf + self.App * xp

        return self.x, b

    def _sensitivity(self, dgdx, dgdb):
        adjoint_load = np.zeros_like(self.x[self.f, ...])

        if dgdx is not None:
            adjoint_load += dgdx[self.f, ...]
        if dgdb is not None:
            adjoint_load += self.Afp * dgdb[self.p, ...]

        lam = np.zeros_like(self.x)
        lamf = -1.0 * self.module_LinSolve.solver.adjoint(adjoint_load)
        lam[self.f, ...] = lamf

        if dgdb is not None:
            lam[self.p, ...] = dgdb[self.p, ...]

        # sensitivities to system matrix
        if self.x.ndim > 1:
            dgdA = DyadCarrier(list(lam.T), list(self.x.T))
        else:
            dgdA = DyadCarrier(lam, self.x)

        # sensitivities to applied load and prescribed state
        dgdbf = np.zeros_like(adjoint_load)
        dgdup = np.zeros_like(self.x[self.p, ...])
        dgdbf -= lam[self.f, ...]
        dgdup += self.Afp.T * lam[self.f, ...]

        if dgdx is not None:
            dgdup += dgdx[self.p, ...]

        if dgdb is not None:
            dgdbf += dgdb[self.f, ...]
            dgdup += self.App * dgdb[self.p, ...]

        return dgdA, dgdbf, dgdup


class Inverse(Module):
    r""" Calculate the inverse of a matrix :math:`\mathbf{B} = \mathbf{A}^{-1}`

    Input Signal:
      - ``A`` (`np.ndarray`): Dense matrix:math:`\mathbf{A}`

    Output Signal:
      - `B` (`np.ndarray`): The inverse of :math:`\mathbf{A}` as dense matrix
    """
    def _response(self, A):
        return np.linalg.inv(A)

    def _sensitivity(self, dB):
        A = self.sig_in[0].state
        B = self.sig_out[0].state
        dA = - B.T @ dB @ B.T
        return dA if np.iscomplexobj(A) else np.real(dA)


# flake8: noqa: C901
def auto_determine_solver(A, isdiagonal=None, islowertriangular=None, isuppertriangular=None,
                          ishermitian=None, issymmetric=None, ispositivedefinite=None):
    """
    Uses parts of Matlab's scheme https://nl.mathworks.com/help/matlab/ref/mldivide.html
    :param A: The matrix
    :param isdiagonal: Manual override for diagonal matrix
    :param islowertriangular: Override for lower triangular matrix
    :param isuppertriangular: Override for upper triangular matrix
    :param ishermitian: Override for hermitian matrix (prevents check)
    :param issymmetric: Override for symmetric matrix (prevents check). Is the same as hermitian for a real matrix
    :param ispositivedefinite: Manual override for positive definiteness
    :return: LinearSolver which should be 'best' for the matrix
    """
    issparse = sps.issparse(A)  # Check if the matrix is sparse
    issquare = A.shape[0] == A.shape[1]  # Check if the matrix is square

    if not issquare:
        if issparse:
            sps.SparseEfficiencyWarning("Only a dense version of QR solver is available")  # TODO
        return SolverDenseQR()

    # l_bw, u_bw = spla.bandwidth(A) # TODO Get bandwidth (implemented in scipy version > 1.8.0)

    if isdiagonal is None:  # Check if matrix is diagonal
        # TODO: This could be improved to check other sparse matrix types as well
        isdiagonal = matrix_is_diagonal(A)
    if isdiagonal:
        return SolverDiagonal()

    # Check if the matrix is triangular
    # TODO Currently only for dense matrices
    if islowertriangular is None:  # Check if matrix is lower triangular
        islowertriangular = False if issparse else np.allclose(A, np.tril(A))
    if islowertriangular:
        warnings.WarningMessage("Lower triangular solver not implemented",
                                UserWarning, getframeinfo(currentframe()).filename, getframeinfo(currentframe()).lineno)

    if isuppertriangular is None:  # Check if matrix is upper triangular
        isuppertriangular = False if issparse else np.allclose(A, np.triu(A))
    if isuppertriangular:
        warnings.WarningMessage("Upper triangular solver not implemented",
                                UserWarning, getframeinfo(currentframe()).filename, getframeinfo(currentframe()).lineno)

    ispermutedtriangular = False
    if ispermutedtriangular:
        warnings.WarningMessage("Permuted triangular solver not implemented",
                                UserWarning, getframeinfo(currentframe()).filename, getframeinfo(currentframe()).lineno)

    # Check if the matrix is complex-valued
    iscomplex = np.iscomplexobj(A)
    if iscomplex:
        # Detect if the matrix is hermitian and/or symmetric
        if ishermitian is None:
            ishermitian = matrix_is_hermitian(A)
        if issymmetric is None:
            issymmetric = matrix_is_symmetric(A)
    else:
        if ishermitian is None and issymmetric is None:
            # Detect if the matrix is symmetric
            issymmetric = matrix_is_symmetric(A)
            ishermitian = issymmetric
        elif ishermitian is not None and issymmetric is not None:
            assert ishermitian == issymmetric, "For real-valued matrices, symmetry and hermitian must be equal"
        elif ishermitian is None:
            ishermitian = issymmetric
        elif issymmetric is None:
            issymmetric = ishermitian

    if issparse:
        # Prefer Intel Pardiso solver as it can solve any matrix TODO: Check for complex matrix
        if SolverSparsePardiso.defined and not iscomplex:
            # TODO check for positive definiteness?  np.all(A.diagonal() > 0) or np.all(A.diagonal() < 0)
            return SolverSparsePardiso(symmetric=issymmetric, hermitian=ishermitian, positive_definite=ispositivedefinite)

        if ishermitian:
            # Check if diagonal is all positive or all negative -> Cholesky
            if np.all(A.diagonal() > 0) or np.all(A.diagonal() < 0):  # TODO what about the complex case?
                if SolverSparseCholeskyScikit.defined:
                    return SolverSparseCholeskyScikit()
                if SolverSparseCholeskyCVXOPT.defined:
                    return SolverSparseCholeskyCVXOPT()

        return SolverSparseLU()  # Default to LU, which should be possible for any non-singular square matrix

    else:  # Dense branch
        if ishermitian:
            # Check if diagonal is all positive or all negative
            if np.all(A.diagonal() > 0) or np.all(A.diagonal() < 0):
                return SolverDenseCholesky()
            else:
                return SolverDenseLDL(hermitian=ishermitian)
        elif issymmetric:
            return SolverDenseLDL(hermitian=ishermitian)
        else:
            # TODO: Detect if the matrix is Hessenberg
            return SolverDenseLU()


class LinSolve(Module):
    r""" Solves linear system of equations :math:`\mathbf{A}\mathbf{x}=\mathbf{b}`

    Self-adjointness is automatically detected using :class:`LDAWrapper`.

    Input Signals:
      - ``A`` (`dense or sparse matrix`): The system matrix :math:`\mathbf{A}` of size ``(n, n)``
      - ``b`` (`vector`): Right-hand-side vector of size ``(n)`` or block-vector of size ``(n, Nrhs)``

    Output Signal:
      - ``x`` (`vector`): Solution vector of size ``(n)`` or block-vector of size ``(n, Nrhs)``

    Keyword Args:
        dep_tol: Tolerance for detecting linear dependence of solution vectors (default = ``1e-5``)
        hermitian: Flag to omit the automatic detection for Hermitian matrix, saves some work for large matrices
        symmetric: Flag to omit the automatic detection for symmetric matrix, saves some work for large matrices
        solver: Manually override the LinearSolver used, instead of the the solver from :func:`auto_determine_solver`

    Attributes:
        use_lda_solver: Use the linear-dependency-aware solver :class:`LDAWrapper` to prevent redundant computations
    """

    use_lda_solver = True

    def _prepare(self, dep_tol=1e-5, hermitian=None, symmetric=None, solver=None):
        self.dep_tol = dep_tol
        self.ishermitian = hermitian
        self.issymmetric = symmetric
        self.solver = solver

    def _response(self, mat, rhs):
        # Do some detections on the matrix type
        self.issparse = sps.issparse(mat)  # Check if it is a sparse matrix
        self.iscomplex = np.iscomplexobj(mat)  # Check if it is a complex-valued matrix
        if not self.iscomplex and self.issymmetric is not None:
            self.ishermitian = self.issymmetric
        if self.ishermitian is None:
            self.ishermitian = matrix_is_hermitian(mat)
        if self.issparse and not self.iscomplex and np.iscomplexobj(rhs):
            raise TypeError("Complex right-hand-side for a real-valued sparse matrix is not supported."
                            "This case can simply be solved by running two rhs (one for the real part and "
                            "one for the imaginary.")

        # Determine the solver we want to use
        if self.solver is None:
            self.solver = auto_determine_solver(mat, ishermitian=self.ishermitian)
        if not isinstance(self.solver, LDAWrapper) and self.use_lda_solver:
            self.solver = LDAWrapper(self.solver, hermitian=self.ishermitian, symmetric=self.issymmetric)

        # Update solver with new matrix
        self.solver.update(mat)

        # Solution
        self.u = self.solver.solve(rhs)

        return self.u

    def _sensitivity(self, dfdv):
        mat, rhs = [s.state for s in self.sig_in]
        lam = self.solver.adjoint(dfdv.conj()).conj()

        if self.issparse:
            if self.u.ndim > 1:
                dmat = DyadCarrier(list(-lam.T), list(self.u.T))
            else:
                dmat = DyadCarrier(-lam, self.u)
        else:
            if self.u.ndim > 1:
                dmat = np.einsum("iB,jB->ij", -lam, self.u, optimize=True)
            else:
                dmat = np.outer(-lam, self.u)
        if not self.iscomplex:
            dmat = dmat.real

        db = np.real(lam) if np.isrealobj(rhs) else lam

        return dmat, db


class EigenSolve(Module):
    r""" Solves the (generalized) eigenvalue problem :math:`\mathbf{A}\mathbf{q}_i = \lambda_i \mathbf{B} \mathbf{q}_i`

    Solves the eigenvalue problem :math:`\mathbf{A}\mathbf{q}_i = \lambda_i \mathbf{q}_i` with normalization
    :math:`\mathbf{q}_i^\text{T} \mathbf{q}_i = 1` or
    the generalized eigenvalue problem :math:`\mathbf{A}\mathbf{q}_i = \lambda_i \mathbf{B} \mathbf{q}_i`
    with normalization :math:`\mathbf{q}_i^\text{T} \mathbf{B} \mathbf{q}_i = 1` (:math:`\mathbf{B}` must be positive
    definite). The eigenvectors are returned as a dense matrix ``Q``, where eigenvector ``Q[:, i]`` belongs to
    eigenvalue ``λ[i]``.

    Both real and complex matrices are supported.
    Correct sensitivities for cases with eigenvalue multiplicity are not implemented yet.

    Mode tracking algorithms can be implemented by the user by providing the argument ``sorting_func``, which is a
    function with arguments (``λ``, ``Q``).

    Todo:
        Support for sparse matrix

    Input Signal(s):
      - ``A`` (`dense matrix`): The system matrix of size ``(n, n)``
      - ``B`` (`dense matrix, optional`): Second system matrix (must be positive-definite) of size ``(n, n)``

    Output Signals:
      - ``λ`` (`vector`): Vector with eigenvalues of size ``(n)``
      - ``Q`` (`matrix`): Matrix with eigenvectors ``Q[:, i]`` corresponding to ``λ[i]``, of size ``(n, n)``

    Keyword Args:
        sorting_func: Sorting function to sort the eigenvalues, which must have signature ``func(λ,Q)``
          (default = ``numpy.argsort``)
        hermitian: Flag to omit the automatic detection for Hermitian matrix, saves some work for large matrices
        nmodes: Number of modes to calculate (only for sparse matrices, default = ``0``)
        sigma: Shift value for the eigenvalue problem (only for sparse matrices). Eigenvalues around the shift are
          calculated first (default = ``0.0``)
    """
    def _prepare(self, sorting_func=lambda W, Q: np.argsort(W), hermitian=None, nmodes=None, sigma=None):
        self.sorting_fn = sorting_func
        self.is_hermitian = hermitian
        self.nmodes = nmodes
        self.sigma = sigma
        self.Ainv = None

    def _response(self, A, *args):
        B = args[0] if len(args) > 0 else None
        if self.is_hermitian is None:
            self.is_hermitian = (matrix_is_hermitian(A) and (B is None or matrix_is_hermitian(B)))
        self.is_sparse = sps.issparse(A) and (B is None or sps.issparse(B))

        # Solve the eigenvalue problem
        if self.is_sparse:
            W, Q = self._sparse_eigs(A, B=B)
        else:
            W, Q = spla.eigh(A, b=B) if self.is_hermitian else spla.eig(A, b=B)

        # Sort the eigenvalues
        isort = self.sorting_fn(W, Q)
        W = W[isort]
        Q = Q[:, isort]

        # Normalize the eigenvectors
        for i in range(W.size):
            qi, wi = Q[:, i], W[i]
            qi *= np.sign(np.real(qi[np.argmax(abs(qi) > 0)]))  # Set first value positive for orientation
            Bqi = qi if B is None else B@qi
            qi /= np.sqrt(qi@Bqi)  # Normalize

            eigvec_nrm = abs(qi@(qi if B is None else B@qi) - 1.0)
            if eigvec_nrm > 1e-5:
                warnings.warn(f"Eigenvector {i} normalization error large: |v^T{'B'if len(args)>0 else ''}v|-1 = {eigvec_nrm}")
            resi_nrm = np.linalg.norm(A@qi - wi*(qi if B is None else B@qi)) / np.linalg.norm(A@qi)
            if resi_nrm > 1e-5:
                warnings.warn(f"Eigenvalue {i} residual large: |Av - λ{'B'if len(args)>0 else ''}v|/|Av| = {resi_nrm}")
        return W, Q

    def _sensitivity(self, dW, dQ):
        A = self.sig_in[0].state
        B = self.sig_in[1].state if len(self.sig_in) > 1 else np.eye(*A.shape)
        dA, dB = None, None
        if not self.is_sparse:
            dA, dB = self._dense_sens(A, B, dW, dQ)
        else:
            if dQ is not None:
                raise NotImplementedError('Sparse eigenvector sensitivities not implemented')
            elif dW is not None:
                dA, dB = self._sparse_eigval_sens(A, B, dW)

        if len(self.sig_in) == 1:
            return dA
        elif len(self.sig_in) == 2:
            return dA, dB


    def _sparse_eigs(self, A, B=None):
        if self.nmodes is None:
            self.nmodes = 6
        if self.sigma is None:
            self.sigma = 0.0

        if self.sigma == 0.0:  # No shift
            mat_shifted = A
        else:
            if B is None:  # If no B is given, use identity to shift
                B = sps.eye(*A.shape)
            mat_shifted = A - self.sigma * B

        # Use shift-and-invert, so make inverse operator
        if self.Ainv is None:
            self.Ainv = auto_determine_solver(mat_shifted, ishermitian=self.is_hermitian)
        self.Ainv.update(mat_shifted)

        AinvOp = spsla.LinearOperator(mat_shifted.shape, matvec=self.Ainv.solve, rmatvec=self.Ainv.adjoint)

        if self.is_hermitian:
            return spsla.eigsh(A, M=B, k=self.nmodes, OPinv=AinvOp, sigma=self.sigma)
        else:
            # TODO
            raise NotImplementedError('Non-Hermitian sparse matrix not supported')

    def _dense_sens(self, A, B, dW, dQ):
        """ Calculates all (eigenvector and eigenvalue) sensitivities for dense matrix """
        dA = np.zeros_like(A)
        dB = np.zeros_like(B) if len(self.sig_in) > 1 else None
        W, Q = [s.state for s in self.sig_out]
        if dW is None:
            dW = np.zeros_like(W)
        if dQ is None:
            dQ = np.zeros_like(Q)

        for i in range(W.size):
            dqi, dwi = dQ[:, i], dW[i]
            if np.linalg.norm(dqi) == 0 and dwi == 0:
                continue
            qi, wi = Q[:, i], W[i]

            P = np.block([[(A - wi * B).T, -((B + B.T) / 2) @ qi[..., np.newaxis]], [-B @ qi.T, 0]])
            adj = np.linalg.solve(P, np.block([dqi, dwi]))  # Adjoint solve Lee(1999)
            nu = adj[:-1]
            alpha = adj[-1]
            dAi = np.outer(-nu, qi)
            dA += dAi if np.iscomplexobj(A) else np.real(dAi)
            if dB is not None:
                dBi = np.outer(wi * nu + alpha / 2 * qi, qi)
                dB += dBi if np.iscomplexobj(B) else np.real(dBi)
        return dA, dB

    def _sparse_eigval_sens(self, A, B, dW):
        if dW is None:
            return None, None
        W, Q = [s.state for s in self.sig_out]
        dA, dB = DyadCarrier(), None if B is None else DyadCarrier()
        for i in range(W.size):
            wi, dwi = W[i], dW[i]
            if dwi == 0:
                continue
            qi = Q[:, i]
            qmq = qi@qi if B is None else qi @ (B @ qi)
            dA_u = (dwi/qmq) * qi
            if np.isrealobj(A):
                dA += DyadCarrier([np.real(dA_u), -np.imag(dA_u)], [np.real(qi), np.imag(qi)])
            else:
                dA += DyadCarrier(dA_u, qi)

            if dB is not None:
                dB_u = (wi*dwi/qmq) * qi
                if np.isrealobj(B):
                    dB -= DyadCarrier([np.real(dB_u), -np.imag(dB_u)], [np.real(qi), np.imag(qi)])
                else:
                    dB -= DyadCarrier(dB_u, qi)
        return dA, dB




import unittest
import numpy as np
import pymoto as pym


class TestConcat(unittest.TestCase):
    def test_concatsignal(self):
        s1 = pym.Signal('sig1', 0.0)
        s2 = pym.Signal('sig2', np.array([1.0, 2.0, 3.0, 4.0]))
        s3 = pym.Signal('sig3', 5.0)
        s4 = pym.Signal('sig4', np.array([[6.0, 7.0]]))
        s5 = pym.Signal('sig5', np.array(8.0))

        s = pym.Signal('out')
        m = pym.ConcatSignal([s1, s2, s3, s4, s5], s)
        m.response()

        self.assertTrue(np.allclose(s.state, np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0])))

        # Set the state of the concatenated signal
        s.sensitivity = np.array([8.0, 7.0, 6.0, 5.0, 4.0, 3.0, 2.0, 1.0, 0.0])
        m.sensitivity()
        self.assertEqual(s1.sensitivity, 8.0)
        self.assertEqual(type(s1.sensitivity), float)

        self.assertTrue(np.allclose(s2.sensitivity, np.array([7.0, 6.0, 5.0, 4.0])))
        self.assertEqual(s2.sensitivity.shape, (4,))
        self.assertEqual(type(s2.sensitivity), np.ndarray)

        self.assertEqual(s3.sensitivity, 3.0)
        self.assertEqual(type(s3.sensitivity), float)

        self.assertTrue(np.allclose(s4.sensitivity, np.array([[2.0, 1.0]])))
        self.assertEqual(type(s4.sensitivity), np.ndarray)
        self.assertEqual(s4.sensitivity.shape, (1, 2))

        self.assertEqual(s5.sensitivity, np.array(0.0))
        self.assertEqual(s5.sensitivity.shape, ())
        self.assertEqual(type(s5.sensitivity), np.ndarray)


if __name__ == '__main__':
    unittest.main()



__version__ = '1.2.1'

from .common.domain import DomainDefinition

# Imports from common
from .common.dyadcarrier import DyadCarrier
from .common.mma import MMA
from .common.solvers import matrix_is_complex, matrix_is_diagonal, matrix_is_symmetric, matrix_is_hermitian, LinearSolver, LDAWrapper
from .common.solvers_dense import SolverDiagonal, SolverDenseQR, SolverDenseLU, SolverDenseCholesky, SolverDenseLDL
from .common.solvers_sparse import SolverSparsePardiso, SolverSparseLU, SolverSparseCholeskyScikit, SolverSparseCholeskyCVXOPT

# Modular inports
from .core_objects import Signal, Module, Network, make_signals

# Import modules
from .modules.assembly import AssembleGeneral, AssembleStiffness, AssembleMass
from .modules.autodiff import AutoMod
from .modules.complex import MakeComplex, RealPart, ImagPart, ComplexNorm
from .modules.filter import FilterConv, Filter, DensityFilter, OverhangFilter
from .modules.generic import MathGeneral, EinSum, ConcatSignal
from .modules.io import PlotDomain, PlotGraph, PlotIter, WriteToVTI
from .modules.linalg import Inverse, LinSolve, EigenSolve, SystemOfEquations, StaticCondensation
from .modules.scaling import Scaling

# Further helper routines
from .routines import finite_difference, minimize_oc, minimize_mma

__all__ = [
    'Signal', 'Module', 'Network', 'make_signals',
    'finite_difference', 'minimize_oc', 'minimize_mma',
    # Common
    'MMA',
    'DyadCarrier',
    'DomainDefinition',
    'matrix_is_complex', 'matrix_is_diagonal', 'matrix_is_symmetric', 'matrix_is_hermitian',
    'LinearSolver', 'LDAWrapper',
    'SolverDiagonal', 'SolverDenseQR', 'SolverDenseLU', 'SolverDenseCholesky', 'SolverDenseLDL',
    'SolverSparsePardiso', 'SolverSparseLU', 'SolverSparseCholeskyScikit', 'SolverSparseCholeskyCVXOPT',
    # Modules
    "MathGeneral", "EinSum", "ConcatSignal",
    "Inverse", "LinSolve", "EigenSolve", "SystemOfEquations", "StaticCondensation",
    "AssembleGeneral", "AssembleStiffness", "AssembleMass",
    "FilterConv", "Filter", "DensityFilter", "OverhangFilter",
    "PlotDomain", "PlotGraph", "PlotIter", "WriteToVTI",
    "MakeComplex", "RealPart", "ImagPart", "ComplexNorm",
    "AutoMod",
    "Scaling"
]



import unittest
import numpy as np
import pymoto as pym
import numpy.testing as npt

np.random.seed(0)


class TestStaticCondensation(unittest.TestCase):
    def test_real_symmetric(self):
        """ Test symmetric real sparse matrix (compliance in 2D)"""
        N = 20
        # Set up the domain
        domain = pym.DomainDefinition(N, N)

        # node groups
        nodes_left = domain.get_nodenumber(0, np.arange(N + 1))
        nodes_right = domain.get_nodenumber(N, np.arange(N + 1))

        # dof groups
        dofs_left = np.repeat(nodes_left * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)
        dofs_right = np.repeat(nodes_right * 2, 2, axis=-1) + np.tile(np.arange(2), N + 1)

        # free and prescribed dofs
        all_dofs = np.arange(0, 2 * domain.nnodes)
        prescribed_dofs = dofs_right
        main_dofs = dofs_left[0::2]
        free_dofs = np.setdiff1d(all_dofs, np.unique(np.hstack([main_dofs, prescribed_dofs])))

        fn = pym.Network()
        sx = pym.Signal('x', np.random.rand(domain.nel))
        sK = fn.append(pym.AssembleStiffness(sx, pym.Signal('K'), domain))
        su = fn.append(pym.StaticCondensation([sK], free=free_dofs, main=main_dofs))
        sc = fn.append(pym.EinSum([su], expression='ij->'))
        fn.response()

        # Check result
        fn1 = pym.Network()
        sK1 = fn1.append(pym.AssembleStiffness(sx, domain=domain, bc=np.concatenate([main_dofs[1:], prescribed_dofs])))
        sf = pym.Signal('f', np.zeros(domain.nnodes*2))
        sf.state[main_dofs[0]] = 1.0
        su1 = fn1.append(pym.LinSolve([sK1, sf]))
        suKu1 = fn1.append(pym.EinSum([su1, sf], expression='i,i->'))
        fn1.response()

        npt.assert_allclose(suKu1.state, 1/su.state[0, 0])

        def tfn(x0, dx, df_an, df_fd): self.assertTrue(np.allclose(df_an, df_fd, rtol=1e-3, atol=1e-5))

        pym.finite_difference(fn, [sx], sc, test_fn=tfn, dx=1e-5, tol=1e-4, verbose=False)



if __name__ == '__main__':
    unittest.main()



from typing import Union, List, Any
import warnings
import inspect
import time
from .utils import _parse_to_list, _concatenate_to_array, _split_from_array
from abc import ABC, abstractmethod


# Local helper functions
def err_fmt(*args):
    """ Format error strings for locating Modules and Signals"""
    err_str = ""
    for a in args:
        err_str += f"\n\t[ {a} ]"
    return err_str


def colored(r, g, b, text):
    """ Colored console text """
    return "\033[38;2;{};{};{}m{} \033[0;0m".format(r, g, b, text)


def stderr_warning(text):
    """ Issue a (colored) warning """
    filename, line, func = get_init_loc()
    warnings.warn_explicit(colored(128, 128, 0, "\nWarning: " + text), RuntimeWarning, filename, line)


def get_init_loc():
    """ Get the location (outside of this file) where the 'current' function is called """
    stk = inspect.stack()
    frame = None
    for fr in stk:
        if fr[1] != __file__:
            frame = fr
            break
    if frame is None:
        return "N/A", "N/A", "N/A"
    _, filename, line, func, _, _ = frame
    return filename, line, func


def get_init_str():
    filename, line, func = get_init_loc()
    return f"File \"{filename}\", line {line}, in {func}"


def fmt_slice(sl):
    """ Formats slices as string
    :param sl: Generic slice or tuple of slices
    :return: Slice(s) formatted as string
    """
    if isinstance(sl, tuple):
        return ", ".join([fmt_slice(sli) for sli in sl])
    elif isinstance(sl, slice):
        return f"{'' if sl.start is None else sl.start}:" \
               f"{'' if sl.stop is None else sl.stop}:" \
               f"{'' if sl.step is None else sl.step}"
    elif hasattr(sl, 'size') and sl.size > 4:
        ndim = sl.ndim if hasattr(sl, 'ndim') else 1
        return "["*ndim + "..." + "]"*ndim
    else:
        return str(sl)


class Signal:
    """
    Manages the state data, sensitivities, and connects module in- and outputs

    Initialize using Signal()
    Optional arguments: tag (string)
    Optional keyword arguments: tag=(string)

    >> Signal('x1')

    >> Signal(tag='x2')

    """
    def __init__(self, tag: str = "", state: Any = None, sensitivity: Any = None):
        """

        :param tag: The name of the signal (string)
        :param state: The initialized state (optional)
        :param sensitivity: The initialized sensitivity (optional)
        """
        self.tag = tag
        self.state = state
        self.sensitivity = sensitivity
        self.keep_alloc = sensitivity is not None

        # Save error string to location where it is initialized
        self._init_loc = get_init_str()

    def _err_str(self):
        return err_fmt(f"Signal \'{self.tag}\', initialized in {self._init_loc}")

    def add_sensitivity(self, ds: Any):
        try:
            if ds is None:
                return
            if self.sensitivity is None:
                self.sensitivity = ds
            else:
                self.sensitivity += ds
            return self
        except TypeError:
            if isinstance(ds, type(self.sensitivity)):
                raise TypeError(f"Cannot add to the sensitivity with type '{type(self.sensitivity).__name__}'"+self._err_str())
            else:
                raise TypeError(f"Adding wrong type '{type(ds).__name__}' to the sensitivity '{type(self.sensitivity).__name__}'"+self._err_str())
        except ValueError:
            sens_shape = self.sensitivity.shape if hasattr(self.sensitivity, 'shape') else ()
            ds_shape = ds.shape if hasattr(ds, 'shape') else ()
            raise ValueError(f"Cannot add argument of shape {ds_shape} to the sensitivity of shape {sens_shape}"+self._err_str()) from None

    def reset(self, keep_alloc: bool = None):
        """ Reset the sensitivities to zero or None
        This must be called to clear internal memory of subsequent sensitivity calculations.
        :param keep_alloc: Keep the sensitivity allocation intact?
        :return: self
        """
        if self.sensitivity is None:
            return self
        if keep_alloc is None:
            keep_alloc = self.keep_alloc
        if keep_alloc:
            try:
                try:
                    self.sensitivity[...] = 0
                except TypeError:
                    self.sensitivity *= 0
            except TypeError:
                stderr_warning(f"reset() - Cannot keep allocation because the operands *= or [] are not defined for sensitivity type \'{type(self.sensitivity).__name__}\'" + self._err_str())
                self.sensitivity = None
        else:
            self.sensitivity = None
        return self

    def __getitem__(self, item):
        """ Obtain a sliced signal, for using its partial contents.
        :param item: Slice indices
        :return: Sliced signal (SignalSlice)
        """
        return SignalSlice(self, item)


class SignalSlice(Signal):
    """ Slice operator for a Signal
    The sliced values are referenced to their original source Signal, such that they can be used and updated in modules.
    This means that updating the values in this SignalSlice changes the data in its source Signal.
    """
    def __init__(self, orig_signal, sl, tag=None):
        self.orig_signal = orig_signal
        self.slice = sl
        self.keep_alloc = False  # Allocation must be False because sensitivity cannot be assigned with [] operator

        # for s in slice:
        if tag is None:
            self.tag = f"{self.orig_signal.tag}[{fmt_slice(self.slice)}]"
        else:
            self.tag = tag

        # Save error string to location where it is initialized
        self._init_loc = get_init_str()

    @property
    def state(self):
        try:
            return None if self.orig_signal.state is None else self.orig_signal.state[self.slice]
        except Exception as e:
            # Possibilities: Unslicable object (TypeError) or Wrong dimensions or out of range (IndexError)
            raise type(e)("SignalSlice.state (getter)" + self._err_str()) from e

    @state.setter
    def state(self, new_state):
        try:
            self.orig_signal.state[self.slice] = new_state
        except Exception as e:
            # Possibilities: Unslicable object (TypeError) or Wrong dimensions or out of range (IndexError)
            raise type(e)("SignalSlice.state (setter)" + self._err_str()) from e

    @property
    def sensitivity(self):
        try:
            return None if self.orig_signal.sensitivity is None else self.orig_signal.sensitivity[self.slice]
        except Exception as e:
            # Possibilities: Unslicable object (TypeError) or Wrong dimensions or out of range (IndexError)
            raise type(e)("SignalSlice.sensitivity (getter)" + self._err_str()) from e

    @sensitivity.setter
    def sensitivity(self, new_sens):
        try:
            if self.orig_signal.sensitivity is None:
                if new_sens is None:
                    return  # Sensitivity doesn't need to be initialized when it is set to None
                try:
                    self.orig_signal.sensitivity = self.orig_signal.state * 0  # Make a new copy with 0 values
                except TypeError:
                    if self.orig_signal.state is None:
                        raise TypeError("Could not initialize sensitivity because state is not set" + self._err_str())
                    else:
                        raise TypeError(f"Could not initialize sensitivity for type \'{type(self.orig_signal.state).__name__}\'")

            if new_sens is None:
                new_sens = 0  # reset() uses this

            self.orig_signal.sensitivity[self.slice] = new_sens
        except Exception as e:
            # Possibilities: Unslicable object (TypeError) or Wrong dimensions or out of range (IndexError)
            raise type(e)("SignalSlice.sensitivity (setter)" + self._err_str()) from e

    def reset(self, keep_alloc: bool = None):
        """ Reset the sensitivities to zero or None
        This must be called to clear internal memory of subsequent sensitivity calculations.
        :param keep_alloc: Keep the sensitivity allocation intact?
        :return: self
        """
        if self.sensitivity is not None:
            self.sensitivity = None
        return self


def make_signals(*args):
    """ Batch-initialize a number of Signals
    :param args: Tags for a number of Signals
    :return: Dictionary of Signals, with key index equal to the signal tag
    """
    ret = dict()
    for a in args:
        ret[a] = Signal(a)
    return ret


def _check_valid_signal(sig: Any):
    """ Checks if the argument is a valid Signal object
    :param sig: The object to check
    :return: True if it is a valid Signal
    """
    if isinstance(sig, Signal):
        return True
    if all([hasattr(sig, f) for f in ["state", "sensitivity", "add_sensitivity", "reset"]]):
        return True
    raise TypeError(f"Given argument with type \'{type(sig).__name__}\' is not a valid Signal")


def _check_valid_module(mod: Any):
    """ Checks if the argument is a valid Module object
    :param mod: The object to check
    :return: True if it is a valid Module
    """
    if issubclass(type(mod), Module):
        return True
    if hasattr(mod, "response") and hasattr(mod, "sensitivity") and hasattr(mod, "reset"):
        return True
    raise TypeError(f"Given argument with type \'{type(mod).__name__}\' is not a valid Module")


def _check_function_signature(fn, signals):
    """ Checks the function signature against given signal list
    :param fn: response_ or sensitivity_ function of a Module
    :param signals: The signals involved
    """
    min_args, max_args = 0, 0
    callstr = f"{type(fn.__self__).__name__}.{fn.__name__}{inspect.signature(fn)}"
    for s, p in inspect.signature(fn).parameters.items():
        if p.kind == inspect.Parameter.POSITIONAL_ONLY or p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
            if p.default != inspect.Parameter.empty:
                raise SyntaxError(f"{callstr} must not have default values \"{p}\"")
            min_args += 1
            if max_args >= 0:
                max_args += 1
        elif p.kind == inspect.Parameter.VAR_POSITIONAL:
            max_args = -1
        elif p.kind == inspect.Parameter.KEYWORD_ONLY:
            raise SyntaxError(f"{callstr} may not contain keyword arguments \"{p}\"")
        elif p.kind == inspect.Parameter.VAR_KEYWORD:
            raise SyntaxError(f"{callstr} may not contain \"{p}\"")
    if len(signals) < min_args:
        raise TypeError(f"Not enough arguments ({len(signals)}) for {callstr}")
    if max_args >= 0 and len(signals) > max_args:
        raise TypeError(f"Too many arguments ({len(signals)}) for {callstr}")


class RegisteredClass(object):
    """
    Abstract base class that can keep track of its subclasses and can instantiate them as well, based on their name.
    """

    @classmethod
    def create(cls, sub_type: str, *args, **kwargs):
        """
        Factory method to create subclasses. Call with the name of the subclass to instantiate a new object of
        requested type.
        :param sub_type: String identifying the name of the subclass (case insensitive)
        :param args: Passed to the subclass constructor
        :param kwargs: Passed to the subclass constructor
        :return: New subclass object
        """
        id_req = sub_type.lower()
        subs = cls.all_subclasses()
        if id_req not in subs.keys():
            raise ValueError("Subclass type not defined: {}".format(sub_type))
        return subs[id_req](*args, **kwargs)

    @classmethod
    def all_subclasses(cls):
        """
        Looks for subclasses of this class, used in creation
        :return: List of (unique) subclasses
        """

        # Recursive search for subclasses
        def get_subs(cl):
            all_subclasses = []

            for subclass in cl.__subclasses__():
                all_subclasses.append(subclass)
                all_subclasses.extend(get_subs(subclass))

            return all_subclasses

        subs = get_subs(cls)

        # Check for duplicates
        seen = {}
        subs_out = {}
        duplicates = []
        duplicate_obj = []
        for sc in subs:
            scn = sc.__name__.lower()
            if scn not in seen:
                seen[scn] = 1
                subs_out[scn] = sc
            else:
                if seen[scn] > 0:
                    duplicates.append(scn)
                    duplicate_obj.append(sc)
                seen[scn] += 1

        # Emit warning if duplicates are found
        for d, do in zip(duplicates, duplicate_obj):
            warnings.warn("Duplicated module '{}', currently defined as {}, duplicate definition at {}"
                          .format(d, subs_out[d], do), Warning)

        return subs_out

    @classmethod
    def print_children(cls):
        print(": ".join([cls.__name__+" subtypes", ", ".join(cls.all_subclasses().keys())]))


class Module(ABC, RegisteredClass):
    """
    Main class: Module
    Transforms input signal to output signal and output signal sensitivity to input signal sensitivity

    Initialize using Module()
    >> Module(input, output)

    Multiple inputs and outputs:
    >> Module([input1, input2], [output1, output2])

    No tag:
    >> Module([input1, input2], [output1, output2])

    No outputs:
    >> Module([inputs])

    Using keywords:
    >> Module(sig_in=[inputs], sig_out=[outputs]
    """

    def _err_str(self, init: bool = True, add_signal: bool = True, fn=None):
        str_list = []
        if init:
            str_list.append(f"Module \'{type(self).__name__}\', initialized in {self._init_loc}")
        if add_signal:
            inp_str = "Inputs: " + ", ".join([s.tag if hasattr(s, 'tag') else 'N/A' for s in self.sig_in]) if len(self.sig_in) > 0 else "No inputs"
            out_str = "Outputs: " + ", ".join([s.tag if hasattr(s, 'tag') else 'N/A' for s in self.sig_out]) if len(self.sig_out) > 0 else "No outputs"
            str_list.append(inp_str + " --> " + out_str)
        if fn is not None:
            name = f"{fn.__self__.__class__.__name__}.{fn.__name__}{inspect.signature(fn)}"
            lineno = inspect.getsourcelines(fn)[1]
            filename = inspect.getfile(fn)
            str_list.append(f"Implemented in File \"{filename}\", line {lineno}, in {name}")
        return err_fmt(*str_list)

    # flake8: noqa: C901
    def __init__(self, sig_in: Union[Signal, List[Signal]] = None, sig_out: Union[Signal, List[Signal]] = None,
                 *args, **kwargs):
        # TODO: Reduce complexity of this init
        self._init_loc = get_init_str()

        self.sig_in = _parse_to_list(sig_in)
        self.sig_out = _parse_to_list(sig_out)
        for i, s in enumerate(self.sig_in):
            try:
                _check_valid_signal(s)
            except Exception as e:
                earg0 = e.args[0] if len(e.args) > 0 else ''
                earg1 = e.args[1:] if len(e.args) > 1 else ()
                raise type(e)(f"Invalid input signal #{i+1} - " + str(earg0) + self._err_str(), *earg1) from None

        for i, s in enumerate(self.sig_out):
            try:
                _check_valid_signal(s)
            except Exception as e:
                earg0 = e.args[0] if len(e.args) > 0 else ''
                earg1 = e.args[1:] if len(e.args) > 1 else ()
                raise type(e)(f"Invalid output signal #{i+1} - " + str(earg0) + self._err_str(), *earg1) from None

        try:
            # Call preparation of submodule with remaining arguments
            self._prepare(*args, **kwargs)
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)("_prepare() - " + str(earg0) + self._err_str(fn=self._prepare), *earg1) from e

        try:
            # Check if the signals match _response() signature
            _check_function_signature(self._response, self.sig_in)
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)(str(earg0) + self._err_str(fn=self._response), *earg1) from None

        try:
            # If no output signals are given, but are required, try to initialize them here
            req_args = 0
            for s, p in inspect.signature(self._sensitivity).parameters.items():
                if p.kind == inspect.Parameter.POSITIONAL_ONLY or p.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    if req_args >= 0:
                        req_args += 1
                elif p.kind == inspect.Parameter.VAR_POSITIONAL:
                    req_args = -1
            if len(self.sig_out) == 0 and req_args >= 0 and req_args != len(self.sig_out):
                # Initialize a number of output signals with default names
                self.sig_out = [Signal(f"{type(self).__name__}_output{i}") for i in range(req_args)]

            # Check if signals match _sensitivity() signature
            _check_function_signature(self._sensitivity, self.sig_out)
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)(str(earg0) + self._err_str(fn=self._sensitivity), *earg1) from None

    def response(self):
        """ Calculate the response from sig_in and output this to sig_out """
        try:
            inp = [s.state for s in self.sig_in]
            state_out = _parse_to_list(self._response(*inp))  # Calculate the actual response

            # Check if enough outputs are calculated
            if len(state_out) != len(self.sig_out):
                raise TypeError(f"Number of responses calculated ({len(state_out)}) is unequal to "
                                f"number of output signals ({len(self.sig_out)})")

            # Update the output signals
            for i, val in enumerate(state_out):
                self.sig_out[i].state = val
            return self
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)("response() - " + str(earg0) + self._err_str(fn=self._response), *earg1) from e

    def __call__(self):
        return self.response()

    def sensitivity(self):
        """  Calculate sensitivities using backpropagation

        Based on the sensitivity we get from sig_out, reverse the process and output the new sensitivities to sig_in
        """
        try:
            # Get all sensitivity values of the outputs
            sens_in = [s.sensitivity for s in self.sig_out]

            if len(self.sig_out) > 0 and all([s is None for s in sens_in]):
                return  # If none of the adjoint variables is set

            # Calculate the new sensitivities of the inputs
            sens_out = _parse_to_list(self._sensitivity(*sens_in))

            # Check if enough sensitivities are calculated
            if len(sens_out) != len(self.sig_in):
                raise TypeError(f"Number of sensitivities calculated ({len(sens_out)}) is unequal to "
                                f"number of input signals ({len(self.sig_in)})")

            # Add the sensitivities to the signals
            for i, ds in enumerate(sens_out):
                self.sig_in[i].add_sensitivity(ds)

            return self
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)("sensitivity() - " + str(earg0) + self._err_str(fn=self._sensitivity), *earg1) from e

    def reset(self):
        """ Reset the state of the sensitivities (they are set to zero or to None) """
        try:
            [s.reset() for s in self.sig_out]
            [s.reset() for s in self.sig_in]
            self._reset()
            return self
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)("reset() - " + str(earg0) + self._err_str(fn=self._reset), *earg1) from e

    # METHODS TO BE DEFINED BY USER
    def _prepare(self, *args, **kwargs):
        pass

    @abstractmethod
    def _response(self, *args):
        raise NotImplementedError("No response behavior defined")

    def _sensitivity(self, *args):
        if len(self.sig_out) > 0 and len(self.sig_in) > 0:
            stderr_warning(f"Sensitivity routine is used, but not defined, in {type(self).__name__}")
        return [None for _ in self.sig_in]

    def _reset(self):
        pass


class Network(Module):
    """ Binds multiple Modules together as one Module

    >> Network(module1, module2, ...)

    >> Network([module1, module2, ...])

    >> Network((module1, module2, ...))

    >> Network([ {type="module1", sig_in=[sig1, sig2], sig_out=[sig3]},
                 {type="module2", sig_in=[sig3], sig_out=[sig4]} ])

    """
    def __init__(self, *args, print_timing=False):
        self._init_loc = get_init_str()
        try:
            # Obtain the internal blocks
            self.mods = _parse_to_list(*args)

            # Check if the blocks are initialized, else create them
            for i, b in enumerate(self.mods):
                if isinstance(b, dict):
                    exclude_keys = ['type']
                    b_ex = {k: b[k] for k in set(list(b.keys())) - set(exclude_keys)}
                    self.mods[i] = Module.create(b['type'], **b_ex)

            # Check validity of modules
            [_check_valid_module(m) for m in self.mods]

            # Gather all the input and output signals of the internal blocks
            all_in = set()
            all_out = set()
            [all_in.update(b.sig_in) for b in self.mods]
            [all_out.update(b.sig_out) for b in self.mods]
            in_unique = all_in - all_out

            # Initialize the parent module, with correct inputs and outputs
            super().__init__(list(in_unique), list(all_out))

            self.print_timing = print_timing
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)(str(earg0) + self._err_str(add_signal=False), *earg1) from None

    def timefn(self, fn):
        start_t = time.time()
        fn()
        print(f"Evaluating {fn} took {time.time() - start_t} s")

    def response(self):
        if self.print_timing:
            [self.timefn(b.response) for b in self.mods]
        else:
            [b.response() for b in self.mods]

    def sensitivity(self):
        [b.sensitivity() for b in reversed(self.mods)]

    def reset(self):
        [b.reset() for b in reversed(self.mods)]

    def _response(self, *args):
        pass  # Unused

    def __copy__(self):
        return Network(*self.mods)

    def copy(self):
        return self.__copy__()

    def __len__(self):
        return len(self.mods)

    def __getitem__(self, i):
        return self.mods[i]

    def __iter__(self):
        return iter(self.mods)

    def append(self, *newmods):
        modlist = _parse_to_list(*newmods)

        # Check if the blocks are initialized, else create them
        for i, b in enumerate(modlist):
            try:  # Check validity of modules
                _check_valid_module(b)
            except Exception as e:
                raise type(e)("append() - Trying to append invalid module " + str(e.args[0])
                              + self._err_str(add_signal=False), *e.args[1:]) from None

        # Obtain the internal blocks
        self.mods.extend(modlist)

        # Gather all the input and output signals of the internal blocks
        all_in = set()
        all_out = set()
        [all_in.update(b.sig_in) for b in self.mods]
        [all_out.update(b.sig_out) for b in self.mods]
        in_unique = all_in - all_out

        self.sig_in = _parse_to_list(in_unique)
        try:
            [_check_valid_signal(s) for s in self.sig_in]
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)("append() - Invalid input signals " + str(earg0)
                          + self._err_str(add_signal=False), *earg1) from None
        self.sig_out = _parse_to_list(all_out)
        try:
            [_check_valid_signal(s) for s in self.sig_out]
        except Exception as e:
            earg0 = e.args[0] if len(e.args) > 0 else ''
            earg1 = e.args[1:] if len(e.args) > 1 else ()
            raise type(e)("append() - Invalid output signals " + str(earg0)
                          + self._err_str(add_signal=False), *earg1) from None

        return modlist[-1].sig_out[0] if len(modlist[-1].sig_out) == 1 else modlist[-1].sig_out  # Returns the output signal
